"""Functions help to create data frames"""

from typing import Any, Union, Callable, Mapping, Iterable, Tuple

from pandas import DataFrame
from varname import argname, varname, VarnameException

import pipda
from pipda import register_func, evaluate_expr
from pipda.expression import Expression
from pipda.symbolic import DirectRefItem, DirectRefAttr

from ..core.types import Dtype, is_null
from ..core.defaults import DEFAULT_COLUMN_PREFIX
from ..core.collections import Collection
from ..core.contexts import Context
from ..core.names import repair_names
from ..core.utils import (
    Array,
    to_df,
    length_of,
    recycle_df,
    df_setitem,
    copy_attrs,
    apply_dtypes,
)


@register_func(None, context=Context.EVAL)
def tibble(
    *args: Any,
    _name_repair: Union[str, Callable] = "check_unique",
    _rows: int = None,
    base0_: bool = None,
    dtypes_: Union[Dtype, Mapping[str, Dtype]] = None,
    frame_: int = 1,
    **kwargs: Any,
) -> DataFrame:
    """Constructs a data frame

    Args:
        *args: and
        **kwargs: A set of name-value pairs.
        _name_repair: treatment of problematic column names:
            - "minimal": No name repair or checks, beyond basic existence,
            - "unique": Make sure names are unique and not empty,
            - "check_unique": (default value), no name repair,
                but check they are unique,
            - "universal": Make the names unique and syntactic
            - a function: apply custom name repair
        _rows: Number of rows of a 0-col dataframe when args and kwargs are
            not provided. When args or kwargs are provided, this is ignored.
        base0_: Whether the suffixes of repaired names should be 0-based.
            If not provided, will use `datar.base.get_option('index.base.0')`.

    Returns:
        A constructed dataframe
    """
    args = tuple((arg for arg in args if arg is not None))
    if not args and not kwargs:
        df = DataFrame() if not _rows else DataFrame(index=range(_rows))
        try:
            df.__dfname__ = varname(ignore=pipda)
        except VarnameException:
            pass
        return df

    names = [None] * len(args)
    values = list(args)
    try:
        argnames = argname(
            "*args",
            frame=frame_,
            ignore=pipda,
            func=tibble,
            vars_only=False,
        )
        if len(argnames) != len(args):
            raise VarnameException
    except VarnameException:
        argnames = None

    if argnames:
        for i, value in enumerate(values):
            if isinstance(value, Expression):
                names[i] = argnames[i]
            elif _expand_value(value) is None and not getattr(
                value, "__name__", getattr(value, "name", None)
            ):
                names[i] = argnames[i]

    names.extend(kwargs)
    values.extend(kwargs.values())

    out = zibble(
        names, values, _name_repair=_name_repair, base0_=base0_, dtypes_=dtypes_
    )

    try:
        out.__dfname__ = varname(frame=frame_, ignore=pipda)
    except VarnameException:
        pass

    if not kwargs and len(args) == 1 and isinstance(args[0], DataFrame):
        copy_attrs(out, args[0])

    return out


@register_func(None, context=Context.EVAL)
def tibble_row(
    *args: Any,
    _name_repair: Union[str, Callable] = "check_unique",
    base0_: bool = None,
    dtypes_: Union[Dtype, Mapping[str, Dtype]] = None,
    **kwargs: Any,
) -> DataFrame:
    """Constructs a data frame that is guaranteed to occupy one row.

    Scalar values will be wrapped with `[]`

    Args:
        *args: and
        **kwargs: A set of name-value pairs.
        _name_repair: treatment of problematic column names:
            - "minimal": No name repair or checks, beyond basic existence,
            - "unique": Make sure names are unique and not empty,
            - "check_unique": (default value), no name repair,
                but check they are unique,
            - "universal": Make the names unique and syntactic
            - a function: apply custom name repair
        base0_: Whether the suffixes of repaired names should be 0-based.
            If not provided, will use `datar.base.get_option('index.base.0')`.

    Returns:
        A constructed dataframe
    """
    if not args and not kwargs:
        df = DataFrame(index=[0])  # still one row
    else:
        df = tibble(
            *args, **kwargs, _name_repair=_name_repair, base0_=base0_, frame_=2
        )

    if df.shape[0] > 1:
        raise ValueError("All arguments must be size one, use `[]` to wrap.")
    try:
        df.__dfname__ = varname()
    except VarnameException:  # pragma: no cover
        df.__dfname__ = None

    apply_dtypes(df, dtypes_)
    return df


def tribble(
    *dummies: Any,
    _name_repair: Union[str, Callable] = "minimal",
    base0_: bool = None,
    dtypes_: Union[Dtype, Mapping[str, Dtype]] = None,
) -> DataFrame:
    """Create dataframe using an easier to read row-by-row layout

    Unlike original API that uses formula (`f.col`) to indicate the column
    names, we use `f.col` to indicate them.

    Args:
        *dummies: Arguments specifying the structure of a dataframe
            Variable names should be specified with `f.name`

    Examples:
        >>> tribble(
        >>>     f.colA, f.colB,
        >>>     "a",    1,
        >>>     "b",    2,
        >>>     "c",    3,
        >>> )

    Returns:
        A dataframe
    """
    columns = []
    data = []
    len_data = 0
    for dummy in dummies:
        # columns
        if isinstance(dummy, (DirectRefAttr, DirectRefItem)):
            columns.append(dummy._pipda_ref)
        elif not columns:
            raise ValueError(
                "Must specify at least one column using the `f.<name>` syntax."
            )
        else:
            if not data:
                data.append([])
            if len(data[-1]) < len(columns):
                data[-1].append(dummy)
                len_data += 1
            else:
                data.append([dummy])
                len_data += 1

    if len_data == 0:
        return zibble(
            columns,
            [[]] * len(columns),
            _name_repair=_name_repair,
            base0_=base0_,
            dtypes_=dtypes_,
        )

    if len_data % len(columns) != 0:
        raise ValueError(
            "Data must be rectangular. "
            f"{len_data} cells is not an integer multiple of "
            f"{len(columns)} columns."
        )

    return zibble(
        columns,
        list(zip(*data)),
        _name_repair=_name_repair,
        base0_=base0_,
        dtypes_=dtypes_,
    )


def zibble(
    names: Iterable[str],
    values: Iterable,
    _name_repair: Union[str, Callable] = "minimal",
    base0_: bool = None,
    dtypes_: Union[Dtype, Mapping[str, Dtype]] = None,
) -> DataFrame:
    """Zip version of tibble, where names specify together and so do values.

    So that it is easlier to create data frame with duplicated names.

    The earlier create columns can be recycled by later expressions.
    For example, `zibble(['x', 'y'], [1, f.x])`. You can also use expressions
    to specify names: `zibble([f.x, f.y], [1, f.x])`.

    Note that for `values`, do not use `c()` to quote all the values, as they
    will be flattened.

    Args:
        names: The names of the columns
            If a name is None/NULL/NA, we should expand the corresponding value
            to get the name
        values: The values for the columns
        _name_repair: treatment of problematic column names:
            - "minimal": No name repair or checks, beyond basic existence,
            - "unique": Make sure names are unique and not empty,
            - "check_unique": (default value), no name repair,
                but check they are unique,
            - "universal": Make the names unique and syntactic
            - a function: apply custom name repair
        base0_: Whether the suffixes of repaired names should be 0-based.
            If not provided, will use `datar.base.get_option('index.base.0')`.

    Returns:
        A data frame
    """
    if len(names) != len(values):
        raise ValueError("Lengths of `names` and `values` are not the same.")

    names = Array(
        [
            name._pipda_ref
            if isinstance(DirectRefAttr, DirectRefItem)
            else name
            for name in names
        ],
        dtype=object,
    )

    out = None
    suffix = 0
    for name, value in zip(names, values):
        # Evaluate value if neccessary

        value = evaluate_expr(value, out, Context.EVAL)
        if isinstance(value, Collection):
            value.expand()
        expanded = _expand_value(value)

        if is_null(name) and expanded is None:
            name = getattr(
                value,
                "__name__",
                getattr(value, "name", f"{DEFAULT_COLUMN_PREFIX}{suffix}"),
            )
            out = _process_one_pair(out, name, value)
        elif is_null(name):
            for nam, val in zip(*expanded):
                out = _process_one_pair(out, nam, val)
        elif expanded is None:
            out = _process_one_pair(out, name, value)
        else:
            sub = None
            for nam, val in zip(*expanded):
                sub = _process_one_pair(sub, nam, val)
            out = _process_one_pair(out, name, sub)

        suffix += 1

    if out is None:
        out = DataFrame()
    names = repair_names(out.columns.tolist(), _name_repair, base0_)
    out.columns = names
    apply_dtypes(out, dtypes_)

    return out


# Helpers ----------------------------------------------------
def _process_one_pair(df: DataFrame, name: str, value: Any) -> DataFrame:
    """Process one name-value pair"""
    if value is None:
        return df

    if df is None:
        # if it is expanded, it should be series
        # so the name must be available
        if isinstance(value, DataFrame):
            # value could be a core.DataFrameGroupBy object
            df = DataFrame(value, copy=True)
            df.columns = [f"{name}${col}" for col in df.columns]
            return df
        return to_df(value, name)

    if length_of(df) == 1:
        df, value = recycle_df(df, value)

    return df_setitem(df, name, value, allow_dups=True)


def _expand_value(value: Any) -> Tuple[str, Iterable[Any]]:
    """Expand value to name-value pairs"""
    if isinstance(value, DataFrame):
        df_dict = value.to_dict("series")
        return df_dict, df_dict.values()
    if isinstance(value, dict):
        return value, value.values()
    return None



"""Cumulative functions"""
from typing import Any

import numpy
from pandas import Series
from pipda import register_func

from ..core.utils import register_numpy_func_x, Array
from ..core.contexts import Context
from ..core.types import is_scalar, is_null

from .na import NA


def _ensure_nas_after_na(x: Any) -> Any:
    """Ensure NAs after first NA

    Since `cum<x>()` in R produces NAs after any NA appearance, but
    numpy/pandas just ignores the NAs and continue producing real values.
    """
    if is_scalar(x):
        return x
    na_indexes = numpy.flatnonzero(is_null(x))
    if len(na_indexes) == 0:
        return x
    na_index = na_indexes[0]
    x = Array(x)
    x[na_index:] = NA
    return x


cumsum = register_numpy_func_x(
    "cumsum",
    "cumsum",
    trans_in=_ensure_nas_after_na,
    doc="""Cumulative sum of elements.

Args:
    x: Input array

Returns:
    An array of cumulative sum of elements in x
""",
)

cumprod = register_numpy_func_x(
    "cumprod",
    "cumprod",
    trans_in=_ensure_nas_after_na,
    doc="""Cumulative product of elements.

Args:
    x: Input array

Returns:
    An array of cumulative product of elements in x
""",
)


@register_func(None, context=Context.EVAL)
def cummin(x: Any) -> Any:
    """Cummulative min along elements in x

    Args:
        x: Input array

    Returns:
        An array of cumulative min of elements in x
    """
    if is_scalar(x):
        return x
    x = _ensure_nas_after_na(x)
    return Series(x).cummin()


@register_func(None, context=Context.EVAL)
def cummax(x: Any) -> Any:
    """Cummulative max along elements in x

    Args:
        x: Input array

    Returns:
        An array of cumulative max of elements in x
    """
    if is_scalar(x):
        return x
    x = _ensure_nas_after_na(x)
    return Series(x).cummax()



"""Bessel function family"""

from typing import Callable, Iterable, Mapping, Tuple, Union

from pipda import register_func

from ..core.contexts import Context


def _get_special_func_from_scipy(name: str) -> Callable:
    """Import bessel functions from scipy on the fly

    In order to make scipy dependency optional
    """
    try:
        from scipy import special
    except ImportError as imperr:  # pragma: no cover
        raise ValueError(
            "Bessel functions require scipy to be installed."
        ) from imperr

    return getattr(special, name)


def _register_bessel_function(
    name: str,
    common_fun: str,
    faster_fun: Mapping[Union[int, float, Tuple], str],
    exp_fun: str = None,
    doc: str = "",
) -> Callable:
    """Register bessel function"""

    if exp_fun is None:

        @register_func(None, context=Context.EVAL)
        def bessel_fun(x: Iterable, nu: float) -> Iterable:
            """Bessel function"""
            # use faster version for order 0 and 1
            if nu in faster_fun:
                fun = _get_special_func_from_scipy(faster_fun[nu])
                return fun(x)

            fun = _get_special_func_from_scipy(common_fun)
            return fun(nu, x)

    else:

        @register_func(None, context=Context.EVAL)
        def bessel_fun(
            x: Iterable, nu: float, expon_scaled: bool = False
        ) -> Iterable:
            """Modified bessel function"""
            # use faster version for order 0 and 1
            if (nu, expon_scaled) in faster_fun:
                fun = _get_special_func_from_scipy(
                    faster_fun[(nu, expon_scaled)]
                )
                return fun(x)

            if expon_scaled:
                fun = _get_special_func_from_scipy(exp_fun)
                return fun(nu, x)

            fun = _get_special_func_from_scipy(common_fun)
            return fun(nu, x)

    bessel_fun.__name__ = name
    bessel_fun.__doc__ = doc
    return bessel_fun


bessel_j = _register_bessel_function(
    "bessel_j",
    common_fun="jv",
    faster_fun={0.0: "j0", 1.0: "j1"},
    doc="""Bessel function of first kind

    Args:
        x: An iterable with numeric >= 0
        nu: The order of the bessel function

    Returns:
        Numeric iterable with the values of the corresponding
        Bessel function.
    """,
)

bessel_y = _register_bessel_function(
    "bessel_y",
    common_fun="yv",
    faster_fun={0.0: "y0", 1.0: "y1"},
    doc="""Bessel function of second kind

    Args:
        x: An iterable with numeric >= 0
        nu: The order of the bessel function

    Returns:
        Numeric iterable with the values of the corresponding
        Bessel function.
    """,
)

bessel_i = _register_bessel_function(
    "bessel_i",
    common_fun="iv",
    exp_fun="ive",
    faster_fun={
        (0.0, False): "i0",
        (1.0, False): "i1",
        (0.0, True): "i0e",
        (1.0, True): "i1e",
    },
    doc="""Modified bessel function of first kind

    Args:
        x: An iterable with numeric >= 0
        nu: The order of the bessel function
        expon_scaled: if TRUE, the results are exponentially scaled
            in order to avoid overflow

    Returns:
        Numeric iterable with scaled values of the corresponding
        Bessel function.
    """,
)

bessel_k = _register_bessel_function(
    "bessel_k",
    common_fun="kv",
    exp_fun="kve",
    faster_fun={
        (0.0, False): "k0",
        (1.0, False): "k1",
        (0.0, True): "k0e",
        (1.0, True): "k1e",
    },
    doc="""Modified bessel function of first kind

    Args:
        x: An iterable with numeric >= 0
        nu: The order of the bessel function
        expon_scaled: if TRUE, the results are exponentially scaled
            in order to avoid underflow

    Returns:
        Numeric iterable with scaled values of the corresponding
        Bessel function.
    """,
)



"""Drop rows containing missing values

https://github.com/tidyverse/tidyr/blob/HEAD/R/drop-na.R
"""

from pandas import DataFrame
from pipda import register_verb

from ..core.contexts import Context
from ..core.utils import arg_match, vars_select, reconstruct_tibble


@register_verb(DataFrame, context=Context.SELECT)
def drop_na(
    _data: DataFrame,
    *columns: str,
    how_: str = "any",
    base0_: bool = None,
) -> DataFrame:
    """Drop rows containing missing values

    See https://tidyr.tidyverse.org/reference/drop_na.html

    Args:
        data: A data frame.
        *columns: Columns to inspect for missing values.
        how_: How to select the rows to drop
            - all: All columns of `columns` to be `NA`s
            - any: Any columns of `columns` to be `NA`s
            (tidyr doesn't support this argument)
        base0_: Whether `*columns` are 0-based if given by indexes
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        Dataframe with rows with NAs dropped and indexes dropped
    """
    arg_match(how_, "how_", ["any", "all"])
    all_columns = _data.columns
    if columns:
        columns = vars_select(all_columns, *columns, base0=base0_)
        columns = all_columns[columns]
        out = _data.dropna(subset=columns, how=how_).reset_index(drop=True)
    else:
        out = _data.dropna(how=how_).reset_index(drop=True)

    return reconstruct_tibble(_data, out, keep_rowwise=True)



"""Function from R-base that can be used as verbs"""
from typing import Any, Iterable, List, Mapping, Sequence, Tuple, Union

import numpy
from pandas import Categorical, DataFrame, Series
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.types import ArrayLikeType, IntType, is_scalar
from ..core.utils import Array, arg_match, get_option, position_after


@register_verb(DataFrame, context=Context.EVAL)
def colnames(
    df: DataFrame, new: Sequence[str] = None, _nested: bool = True
) -> Union[List[Any], DataFrame]:
    """Get or set the column names of a dataframe

    Args:
        df: The dataframe
        new: The new names to set as column names for the dataframe.

    Returns:
        A list of column names if names is None, otherwise return the dataframe
        with new column names.
        if the input dataframe is grouped, the structure is kept.
    """
    from ..stats.verbs import set_names

    if not _nested:
        if new is not None:
            return set_names(df, new, __calling_env=CallingEnvs.REGULAR)
        return df.columns.tolist()

    if new is not None:
        namei = 0
        newnames = []
        last_parts0 = None
        for colname in df.columns:
            parts = str(colname).split("$", 1)
            if not newnames:
                if len(parts) < 2:
                    newnames.append(new[namei])
                else:
                    last_parts0 = parts[0]
                    newnames.append(f"{new[namei]}${parts[1]}")
            elif len(parts) < 2:
                namei += 1
                newnames.append(new[namei])
            elif last_parts0 and colname.startswith(f"{last_parts0}$"):
                newnames.append(f"{new[namei]}${parts[1]}")
            else:
                namei += 1
                newnames.append(f"{new[namei]}${parts[1]}")
                last_parts0 = parts[0]
        return set_names(df, newnames, __calling_env=CallingEnvs.REGULAR)

    cols = [
        col.split("$", 1)[0] if isinstance(col, str) else col
        for col in df.columns
    ]
    out = []
    for col in cols:
        if col not in out:
            out.append(col)
    return out


@register_verb(DataFrame, context=Context.EVAL)
def rownames(
    df: DataFrame, new: Sequence[str] = None
) -> Union[List[Any], DataFrame]:
    """Get or set the row names of a dataframe

    Args:
        df: The dataframe
        new: The new names to set as row names for the dataframe.
        copy: Whether return a copy of dataframe with new row names

    Returns:
        A list of row names if names is None, otherwise return the dataframe
        with new row names.
        if the input dataframe is grouped, the structure is kept.
    """
    if new is not None:
        df = df.copy()
        df.index = new
        return df

    return df.index.tolist()


@register_verb(DataFrame, context=Context.EVAL)
def dim(x: DataFrame, _nested: bool = True) -> Tuple[int]:
    """Retrieve the dimension of a dataframe.

    Args:
        x: a dataframe
        _nested: When there is _nesteded df, count as 1.

    Returns:
        The shape of the dataframe.
    """
    return (
        nrow(x, __calling_env=CallingEnvs.REGULAR),
        ncol(x, _nested, __calling_env=CallingEnvs.REGULAR),
    )


@register_verb(DataFrame)
def nrow(_data: DataFrame) -> int:
    """Get the number of rows in a dataframe

    Args:
        _data: The dataframe

    Returns:
        The number of rows in _data
    """
    return _data.shape[0]


@register_verb(DataFrame)
def ncol(_data: DataFrame, _nested: bool = True):
    """Get the number of columns in a dataframe

    Args:
        _data: The dataframe
        _nested: When there is _nesteded df, count as 1.

    Returns:
        The number of columns in _data
    """
    if not _nested:
        return _data.shape[1]
    cols = set()
    for col in _data.columns:
        cols.add(col.split("$", 1)[0] if isinstance(col, str) else col)
    return len(cols)


@register_verb(context=Context.EVAL)
def diag(
    x: Any = 1,
    nrow: IntType = None,
    ncol: IntType = None,
) -> DataFrame:
    """Extract, construct a diagonal dataframe or replace the diagnal of
    a dataframe.

    When used with DataFrameGroupBy data, groups are ignored

    Args:
        x: a matrix, vector or scalar
        nrow, ncol: optional dimensions for the result when x is not a matrix.
            if nrow is an iterable, it will replace the diagnal of the input
            dataframe.

    Returns:
        If x is a matrix then diag(x) returns the diagonal of x.
        In all other cases the value is a diagonal matrix with nrow rows and
        ncol columns (if ncol is not given the matrix is square).
        Here nrow is taken from the argument if specified, otherwise
        inferred from x
    """
    if nrow is None and isinstance(x, int):
        nrow = x
        x = 1
    if ncol is None:
        ncol = nrow
    if is_scalar(x):
        nmax = max(nrow, ncol)
        x = [x] * nmax
    elif nrow is not None:
        nmax = max(nrow, ncol)
        nmax = nmax // len(x)
        x = x * nmax

    x = Array(x)
    ret = DataFrame(numpy.diag(x), dtype=x.dtype)
    return ret.iloc[:nrow, :ncol]


@diag.register(DataFrame)
def _(
    x: DataFrame,
    nrow: Any = None,
    ncol: IntType = None,
) -> Union[DataFrame, numpy.ndarray]:
    """Diag when x is a dataframe"""
    if nrow is not None and ncol is not None:
        raise ValueError("Extra arguments received for diag.")

    x = x.copy()
    if nrow is not None:
        numpy.fill_diagonal(x.values, nrow)
        return x
    return numpy.diag(x)


@register_verb(DataFrame)
def t(_data: DataFrame, copy: bool = False) -> DataFrame:
    """Get the transposed dataframe

    Args:
        _data: The dataframe
        copy: When copy the data in memory

    Returns:
        The transposed dataframe.
    """
    return _data.transpose(copy=copy)


@register_verb(DataFrame)
def names(
    x: DataFrame, new: Sequence[str] = None, _nested: bool = True
) -> Union[List[str], DataFrame]:
    """Get the column names of a dataframe"""
    return colnames(x, new, _nested, __calling_env=CallingEnvs.REGULAR)


@names.register(dict)
def _(
    x: Mapping[str, Any], new: Iterable[str] = None, _nested: bool = True
) -> Union[List[str], Mapping[str, Any]]:
    """Get the keys of a dict
    dict is like a list in R, mimic `names(<list>)` in R.
    """
    if new is None:
        return list(x)
    return dict(zip(new, x.values()))


@register_verb(context=Context.EVAL)
def setdiff(x: Any, y: Any) -> List[Any]:
    """Diff of two iterables"""
    if is_scalar(x):
        x = [x]
    if is_scalar(y):
        y = [y]
    return [elem for elem in x if elem not in y]


@register_verb(context=Context.EVAL)
def intersect(x: Any, y: Any) -> List[Any]:
    """Intersect of two iterables"""
    if is_scalar(x):
        x = [x]
    if is_scalar(y):
        y = [y]
    return [elem for elem in x if elem in y]


@register_verb(context=Context.EVAL)
def union(x: Any, y: Any) -> List[Any]:
    """Union of two iterables"""
    if is_scalar(x):
        x = [x]
    if is_scalar(y):
        y = [y]

    return list(x) + setdiff(y, x, __calling_env=CallingEnvs.REGULAR)


@register_verb(context=Context.EVAL)
def setequal(x: Any, y: Any) -> List[Any]:
    """Check set equality for two iterables (order doesn't matter)"""
    if is_scalar(x):
        x = [x]
    if is_scalar(y):
        y = [y]
    x = sorted(x)
    y = sorted(y)
    return x == y


@register_verb(
    (list, tuple, numpy.ndarray, Series, Categorical),
    context=Context.EVAL,
)
def append(x: Any, values: Any, after: int = -1, base0_: bool = None) -> List:
    """Add elements to a vector.

    Args:
        x: the vector the values are to be appended to.
        values: to be included in the modified vector.
        after: a subscript, after which the values are to be appended.
        base0_: Whether after is 0-based.
            if not given, will use `get_option("index.base.0")`.
            When it's 1-based, after=0 will append to the beginning,
            -1 will append to the end.
            When 0-based, after=None will append to the beginning,
            -1 to the end

    Returns:
        A vector containing the values in ‘x’ with the elements of
        ‘values’ appended after the specified element of ‘x’.
    """
    # if is_scalar(x):
    #     x = [x]
    if is_scalar(values):
        values = [values]
    x = list(x)
    values = list(values)

    base0_ = get_option("index.base.0", base0_)
    # 0 is not allowed with 1-base
    if base0_ and after is None:
        return values + x
    pos = position_after(after, len(x), base0_)
    return x[:pos] + values + x[pos:]


@register_verb((list, tuple, numpy.ndarray, Series, Categorical))
def duplicated(
    x: Iterable[Any],
    incomparables: Sequence[Any] = None,
    from_last: bool = False,
) -> numpy.ndarray:
    """Determine Duplicate Elements

    Args:
        x: The iterable to detect duplicates
            Currently, elements in `x` must be hashable.
        from_last: Whether start to detect from the last element

    Returns:
        A bool array with the same length as `x`
    """
    dups = set()
    out = []
    out_append = out.append
    if incomparables is None:
        incomparables = []

    if from_last:
        x = reversed(x)
    for elem in x:
        if elem in incomparables:
            out_append(False)
        elif elem in dups:
            out_append(True)
        else:
            dups.add(elem)
            out_append(False)
    if from_last:
        out = list(reversed(out))
    return Array(out, dtype=bool)


@duplicated.register(DataFrame)
def _(
    x: DataFrame,
    incomparables: Iterable[Any] = None,
    from_last: bool = False,
) -> numpy.ndarray:
    """Check if rows in a data frame are duplicated

    `incomparables` not working here
    """
    keep = "first" if not from_last else "last"
    return x.duplicated(keep=keep).values


@register_verb(DataFrame)
def max_col(
    df: DataFrame, ties_method: str = "random", base0_: bool = None
) -> Iterable[int]:
    """Find the maximum position for each row of a matrix

    Args:
        df: The data frame
        ties_method: how ties are handled
            - "random": use a random index
            - "first" use the first index
            - "last" use the last index
        base0_: Whether the returned indices are 0-based
            If not provided, will use `get_option("which_base_0")`

    Returns:
        The indices of max values for each row
    """
    ties_method = arg_match(
        ties_method, "ties_method", ["random", "first", "last"]
    )
    base = int(not get_option("which_base_0", base0_))

    def which_max_with_ties(ser: Series):
        """Find index with max if ties happen"""
        indices = numpy.flatnonzero(ser == max(ser)) + base
        if len(indices) == 1 or ties_method == "first":
            return indices[0]
        if ties_method == "random":
            return numpy.random.choice(indices)
        return indices[-1]

    return df.apply(which_max_with_ties, axis=1).to_numpy()


@register_verb(DataFrame)
def complete_cases(_data: DataFrame) -> Iterable[bool]:
    """Return a logical vector indicating values of rows are complete.

    Args:
        _data: The dataframe

    Returns:
        A logical vector specifying which observations/rows have no
        missing values across the entire sequence.
    """
    return _data.apply(lambda row: row.notna().all(), axis=1).values


@register_verb(DataFrame)
def proportions(
    x: DataFrame, margin: Union[int, tuple, list] = None
) -> DataFrame:
    """Returns conditional proportions given `margins` (alias: prop_table)

    Args:
        x: A numeric table
        margin: If x is a dataframe, 1 for rows, 2 for columns, and 3 or [1,2]
            for both rows and columns (turns x to a unit matrix with all
            elements equal to 1)

    Returns:
        The x but with given margin replaced with proportion
    """
    from ..dplyr import mutate, rename_with, across, everything
    from . import sum_

    if margin is None:
        sumall = x.to_numpy().sum()
        return x.applymap(lambda elem: elem / sumall, na_action="ignore")

    if margin == 1:
        index = x.index
        out = t(
            proportions(
                rename_with(
                    t(x, __calling_env=CallingEnvs.REGULAR),
                    str,
                    __calling_env=CallingEnvs.REGULAR,
                ),
                2,
                __calling_env=CallingEnvs.REGULAR,
            ),
            __calling_env=CallingEnvs.REGULAR,
        )
        out.index = index
        return out

    if margin == 2:
        return mutate(
            x,
            across(

                everything(__calling_env=CallingEnvs.PIPING),
                lambda col: col / sum_(col, __calling_env=CallingEnvs.REGULAR),
                __calling_env=CallingEnvs.PIPING,
            ),
            __calling_env=CallingEnvs.REGULAR,
        )

    return x.applymap(lambda elem: 1, na_action="ignore")


prop_table = proportions


@proportions.register((list, tuple, numpy.array, Series))
def _(
    x: ArrayLikeType, margin: Union[int, tuple, list] = None
) -> ArrayLikeType:
    """proportions for vectors"""
    from . import sum_

    if isinstance(x, (list, tuple)):
        x = Array(x)

    return x / sum_(x, __calling_env=CallingEnvs.REGULAR)



"""NA related constants or functions"""

from typing import Any, Iterable

import numpy
from pipda import register_func

from ..core.contexts import Context
from ..core.types import is_null, is_scalar
from ..core.defaults import NA_REPR
from ..core.utils import register_numpy_func_x

NA = numpy.nan
NaN = NA
Inf = numpy.inf

# Just for internal and testing uses
NA_character_ = NA_REPR
NA_integer_ = numpy.random.randint(numpy.iinfo(numpy.int32).max)
NA_real_ = numpy.nan
NA_compex_ = complex(NA_real_, NA_real_)


@register_func(None, context=Context.EVAL)
def is_na(x: Any) -> Iterable[bool]:
    """Test if a value is nullable or elements in the value is nullable

    Args:
        x: The value to test

    Returns:
        If `x` is scalar, returns a scalar bool. Otherwise, return an array
        of bools.
    """
    return is_null(x)


@register_func(None, context=Context.EVAL)
def any_na(x: Any, recursive: bool = False) -> bool:
    """Check if any element in the value is nullable

    Args:
        x: The value to check
        recursive: Whether do a recursive check

    Returns:
        True if any element is nullable otherwise False
    """
    if is_scalar(x):
        return is_na(x)

    from .logical import is_true

    if not recursive:
        return any(is_true(is_na(elem)) for elem in x)

    out = False
    for elem in x:
        if any_na(elem, recursive=True):
            return True
    return out


is_infinite = register_numpy_func_x(
    "is_infinite",
    "isinf",
    doc="""Check if a value or values are infinite numbers

    Args:
        x: The value to check

    Returns:
        True if the value is infinite, False otherwise
        For iterable values, returns the element-wise results
    """
)

is_finite = register_numpy_func_x(
    "is_finite",
    "isfinite",
    doc="""Check if a value or values are finite numbers

    Args:
        x: The value to check

    Returns:
        True if the value is finite, False otherwise
        For iterable values, returns the element-wise results
    """
)

is_nan = register_numpy_func_x(
    "is_nan",
    "isnan",
    doc="""Check if a value or values are NaNs

    Args:
        x: The value to check

    Returns:
        True if the value is nan, False otherwise
        For iterable values, returns the element-wise results
    """
)



"""Operators for datar"""
from typing import Any, Tuple
from functools import partial
import operator

import numpy
from pandas import Series
from pipda import register_operator, Operator

from .utils import length_of, recycle_value
from .collections import Collection, Inverted, Negated, Intersect
from .exceptions import DataUnrecyclable
from .types import BoolOrIter


@register_operator
class DatarOperator(Operator):
    """Operator class for datar"""

    def _arithmetize1(self, operand: Any, op: str) -> Any:
        """Operator for single operand"""
        op_func = getattr(operator, op)
        # Data length might be changed after evaluation
        # operand = recycle_value(operand, self.data.shape[0])
        return op_func(operand)

    def _arithmetize2(self, left: Any, right: Any, op: str) -> Any:
        """Operator for paired operands"""
        op_func = getattr(operator, op)
        left, right = _recycle_left_right(left, right)
        return op_func(left, right)

    def _op_invert(self, operand: Any) -> Any:
        """Interpretation for ~x"""
        if isinstance(operand, (slice, str, list, tuple)):
            return Inverted(operand)
        return self._arithmetize1(operand, "invert")

    def _op_neg(self, operand: Any) -> Any:
        """Interpretation for -x"""
        if isinstance(operand, (slice, list)):
            return Negated(operand)
        return self._arithmetize1(operand, "neg")

    def _op_and_(self, left: Any, right: Any) -> Any:
        """Mimic the & operator in R.

        This has to have Expression objects to be involved to work

        Args:
            left: Left operand
            right: Right operand

        Returns:
            The intersect of the columns
        """
        if isinstance(left, list):
            # induce an intersect with Collection
            return Intersect(left, right)

        left, right = _recycle_left_right(left, right)
        left = Series(left).fillna(False)
        right = Series(right).fillna(False)
        return left & right

    def _op_or_(self, left: Any, right: Any) -> Any:
        """Mimic the & operator in R.

        This has to have Expression objects to be involved to work

        Args:
            left: Left operand
            right: Right operand

        Returns:
            The intersect of the columns
        """
        if isinstance(left, list):
            return Collection(left, right)

        left, right = _recycle_left_right(left, right)
        left = Series(left).fillna(False)
        right = Series(right).fillna(False)
        return left | right

    def _op_ne(self, left: Any, right: Any) -> BoolOrIter:
        """Interpret for left != right"""
        out = self._op_eq(left, right)
        if isinstance(out, (numpy.ndarray, Series)):
            neout = ~out
            # neout[pandas.isna(out)] = numpy.nan
            return neout
        # out is always a numpy.ndarray
        return not out  # pragma: no cover

    def __getattr__(self, name: str) -> Any:
        """Other operators"""
        if name.startswith('_op_'):
            attr = partial(self._arithmetize2, op=name[4:])
            attr.__qualname__ = self._arithmetize2.__qualname__
            return attr
        return super().__getattr__(name)


def _recycle_left_right(left: Any, right: Any) -> Tuple:
    """Recycle left right operands to each other"""
    try:
        left = recycle_value(left, length_of(right))
    except DataUnrecyclable:
        right = recycle_value(right, length_of(left))
    return left, right



"""Utilities for forcats"""
from pandas import Categorical
from ..core.types import ForcatsType, is_categorical, is_scalar


def check_factor(_f: ForcatsType) -> Categorical:
    """Make sure the input become a factor"""
    if not is_categorical(_f):
        if is_scalar(_f):
            _f = [_f]
        return Categorical(_f)

    return _f



"""Expand data frame to include all possible combinations of values

https://github.com/tidyverse/tidyr/blob/HEAD/R/expand.R
"""

from typing import Any, Callable, Iterable, Mapping, Union

import numpy
import pandas
from numpy import product
from pandas import DataFrame, Series, Categorical
from pandas.core.dtypes.common import is_categorical_dtype
from pipda import register_func, register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.defaults import DEFAULT_COLUMN_PREFIX
from ..core.types import is_scalar, is_null, is_not_null
from ..core.utils import categorized, copy_attrs, reconstruct_tibble
from ..core.grouped import DataFrameGroupBy, DataFrameRowwise
from ..core.names import repair_names

from ..base import NA, NULL, factor, levels
from ..tibble import tibble
from ..dplyr import arrange, distinct, pull


@register_func(None, context=Context.EVAL)
def expand_grid(
    *args: Iterable[Any],
    _name_repair: Union[str, Callable] = "check_unique",
    base0_: bool = None,
    **kwargs: Iterable[Any],
) -> DataFrame:
    """Create a tibble from all combinations of inputs

    Args:
        *args: and
        **kwargs: name-value pairs.
            For `*args`, names will be inferred from the values and if failed,
            `_Var0`, `_Var1`, etc will be used.
        _name_repair: treatment of problematic column names:
            - "minimal": No name repair or checks, beyond basic existence,
            - "unique": Make sure names are unique and not empty,
            - "check_unique": (default value), no name repair,
                but check they are unique,
            - "universal": Make the names unique and syntactic
            - a function: apply custom name repair
        base0_: Whether the suffixes of repaired names should be 0-based.
            If not provided, will use `datar.base.get_option('index.base.0')`.

    Returns:
        A data frame with one column for each input in `*args` and `**kwargs`.
        The output will have one row for each combination of the inputs,
        i.e. the size be equal to the product of the sizes of the inputs.
        This implies that if any input has length 0, the output will have
        zero rows.
    """
    dots = _dots_cols(*args, **kwargs)
    named = dots.pop("__named__")
    ns = {key: len(val) for key, val in dots.items()}
    n = product(list(ns.values()))

    if n == 0:
        out = {
            key: (val.iloc[[], :] if isinstance(val, DataFrame) else [])
            for key, val in dots.items()
        }
    else:
        n = numpy.array([n], dtype=float)
        ns_np = numpy.array(list(ns.values()), dtype=float)

        each = n / numpy.cumprod(ns_np)
        times = n / each / ns_np

        each = dict(zip(dots, each.astype(int)))
        times = dict(zip(dots, times.astype(int)))
        out = {
            key: _vec_repeat(val, each[key], times[key])
            for key, val in dots.items()
        }

    # # tibble will somehow flatten the nested dataframes into fake nested df.
    # # do it inside _flatten_nested
    # out = tibble(out, _name_repair=_name_repair, base0_=base0_)
    return _flatten_nested(out, named, _name_repair, base0_)


@register_verb(DataFrame, context=Context.EVAL)
def expand(
    data: DataFrame,
    *args: Union[Series, DataFrame],
    _name_repair: Union[str, Callable] = "check_unique",
    base0_: bool = None,
    **kwargs: Union[Series, DataFrame],
) -> DataFrame:
    """Generates all combination of variables found in a dataset.

    Args:
        data: A data frame
        *args: and,
        **kwargs: columns to expand. Columns can be atomic lists.
            - To find all unique combinations of x, y and z, including
              those not present in the data, supply each variable as a
              separate argument: `expand(df, x, y, z)`.
            - To find only the combinations that occur in the data, use
              nesting: `expand(df, nesting(x, y, z))`.
            - You can combine the two forms. For example,
              `expand(df, nesting(school_id, student_id), date)` would
              produce a row for each present school-student combination
              for all possible dates.
        _name_repair: treatment of problematic column names:
            - "minimal": No name repair or checks, beyond basic existence,
            - "unique": Make sure names are unique and not empty,
            - "check_unique": (default value), no name repair,
                but check they are unique,
            - "universal": Make the names unique and syntactic
            - a function: apply custom name repair
        base0_: Whether the suffixes of repaired names should be 0-based.
            If not provided, will use `datar.base.get_option('index.base.0')`.

    Returns:
        A data frame with all combination of variables.
    """
    cols = _dots_cols(*args, **kwargs)
    named = cols.pop("__named__")
    cols = {key: _sorted_unique(val) for key, val in cols.items()}

    out = expand_grid(**cols, _name_repair=_name_repair, base0_=base0_)
    out = _flatten_nested(out, named, _name_repair, base0_)

    copy_attrs(out, data)
    return out


@expand.register(DataFrameGroupBy, context=Context.PENDING)
def _(
    data: DataFrameGroupBy,
    *args: Union[Series, DataFrame],
    _name_repair: Union[str, Callable] = "check_unique",
    base0_: bool = None,
    **kwargs: Union[Series, DataFrame],
) -> DataFrameGroupBy:
    """Expand on grouped data frame"""

    def apply_func(df):
        return expand(
            df,
            *args,
            _name_repair=_name_repair,
            base0_=base0_,
            **kwargs,
            __calling_env=CallingEnvs.REGULAR,
        )

    out = data._datar_apply(apply_func)
    return reconstruct_tibble(data, out)


@expand.register(DataFrameRowwise, context=Context.EVAL)
def _(
    data: DataFrameRowwise,
    *args: Union[Series, DataFrame],
    _name_repair: Union[str, Callable] = "check_unique",
    base0_: bool = None,
    **kwargs: Union[Series, DataFrame],
) -> DataFrame:
    """Expand on rowwise dataframe"""
    return expand.dispatch(DataFrame)(
        data, *args, _name_repair=_name_repair, base0_=base0_, **kwargs
    )


@register_func(None, context=Context.EVAL)
def nesting(
    *args: Any,
    _name_repair: Union[str, Callable] = "check_unique",
    base0_: bool = None,
    **kwargs: Any,
) -> DataFrame:
    """A helper that only finds combinations already present in the data.

    Args:
        *args: and,
        **kwargs: columns to expand. Columns can be atomic lists.
            - To find all unique combinations of x, y and z, including
              those not present in the data, supply each variable as a
              separate argument: `expand(df, x, y, z)`.
            - To find only the combinations that occur in the data, use
              nesting: `expand(df, nesting(x, y, z))`.
            - You can combine the two forms. For example,
              `expand(df, nesting(school_id, student_id), date)` would
              produce a row for each present school-student combination
              for all possible dates.
        _name_repair: treatment of problematic column names:
            - "minimal": No name repair or checks, beyond basic existence,
            - "unique": Make sure names are unique and not empty,
            - "check_unique": (default value), no name repair,
                but check they are unique,
            - "universal": Make the names unique and syntactic
            - a function: apply custom name repair
        base0_: Whether the suffixes of repaired names should be 0-based.
            If not provided, will use `datar.base.get_option('index.base.0')`.

    Returns:
        A data frame with all combinations in data.
    """
    cols = _dots_cols(*args, **kwargs)
    named = cols.pop("__named__")
    out = _sorted_unique(
        tibble(**cols, _name_repair=_name_repair, base0_=base0_)
    )
    return _flatten_nested(out, named, _name_repair, base0_)


@register_func(None, context=Context.EVAL)
def crossing(
    *args: Any,
    _name_repair: Union[str, Callable] = "check_unique",
    base0_: bool = None,
    **kwargs: Any,
) -> DataFrame:
    """A wrapper around `expand_grid()` that de-duplicates and sorts its inputs

    When values are not specified by literal `list`, they will be sorted.

    Args:
        *args: and,
        **kwargs: columns to expand. Columns can be atomic lists.
            - To find all unique combinations of x, y and z, including
              those not present in the data, supply each variable as a
              separate argument: `expand(df, x, y, z)`.
            - To find only the combinations that occur in the data, use
              nesting: `expand(df, nesting(x, y, z))`.
            - You can combine the two forms. For example,
              `expand(df, nesting(school_id, student_id), date)` would
              produce a row for each present school-student combination
              for all possible dates.
        _name_repair: treatment of problematic column names:
            - "minimal": No name repair or checks, beyond basic existence,
            - "unique": Make sure names are unique and not empty,
            - "check_unique": (default value), no name repair,
                but check they are unique,
            - "universal": Make the names unique and syntactic
            - a function: apply custom name repair
        base0_: Whether the suffixes of repaired names should be 0-based.
            If not provided, will use `datar.base.get_option('index.base.0')`.

    Returns:
        A data frame with values deduplicated and sorted.
    """
    cols = _dots_cols(*args, **kwargs)
    named = cols.pop("__named__")
    out = {key: _sorted_unique(val) for key, val in cols.items()}

    out = expand_grid(**out, _name_repair=_name_repair, base0_=base0_)
    return _flatten_nested(out, named, _name_repair, base0_)


# Helpers --------------------------------
def _dots_cols(
    *args: Iterable[Any], **kwargs: Iterable[Any]
) -> Mapping[str, Iterable[Any]]:
    """Mimic tidyr:::dots_cols to clean up the dot (args, kwargs) arugments"""
    out = {"__named__": {}}
    for i, arg in enumerate(args):
        if arg is None:
            continue

        name = getattr(
            arg,
            "__dfname__",
            getattr(arg, "name", getattr(arg, "__name__", None)),
        )
        name = name or f"{DEFAULT_COLUMN_PREFIX}{i}"
        out["__named__"][name] = False
        out[name] = [arg] if is_scalar(arg) else arg

    for name, arg in kwargs.items():
        if arg is None:
            continue

        out[name] = [arg] if is_scalar(arg) else arg
        out["__named__"][name] = True

    return out


def _vec_repeat(
    vec: Iterable[Any], each: Iterable[int], times: Iterable[int]
) -> Iterable[Any]:
    """Repeat a vector or a dataframe by rows"""
    if isinstance(vec, DataFrame):
        indexes = _vec_repeat(vec.index, each=each, times=times)
        return vec.loc[indexes, :].reset_index(drop=True)

    vec = categorized(vec)
    # numpy.repeat() turn [numpy.nan, 'A'] to ['nan', 'A']
    vec_to_rep = vec
    if any(isinstance(elem, str) for elem in vec) and any(
        pandas.isnull(elem) for elem in vec
    ):
        vec_to_rep = numpy.array(vec, dtype=object)
    out = numpy.tile(numpy.repeat(vec_to_rep, each), times)
    if is_categorical_dtype(vec):
        return factor(out, levels(vec), ordered=vec.ordered)
    return out


def _flatten_nested(
    x: Union[DataFrame, Mapping[str, Iterable[Any]]],
    named: Mapping[str, bool],
    name_repair: Union[str, Callable],
    base0: bool = None,
) -> DataFrame:
    """Mimic `tidyr:::flatten_nested`"""
    if isinstance(x, DataFrame):
        names = repair_names(list(named), name_repair, base0)
        named = dict(zip(names, named.values()))
        x = {
            name: pull(x, name, __calling_env=CallingEnvs.REGULAR)
            for name in named
        }

    to_flatten = {
        key: isinstance(val, DataFrame) and not named[key]
        for key, val in x.items()
    }
    out = _flatten_at(x, to_flatten)
    return tibble(**out, _name_repair=name_repair, base0_=base0)


def _flatten_at(
    x: Mapping[str, Iterable[Any]], to_flatten: Mapping[str, bool]
) -> Mapping[str, Iterable[Any]]:
    """Flatten data at `to_flatten`"""
    if not any(to_flatten.values()):
        return x

    out = {}
    for name, val in x.items():
        if len(val) == 0:
            continue

        if to_flatten[name]:
            for col in val:
                out[col] = val[col]
        else:
            out[name] = val
    return out


def _sorted_unique(x: Iterable[Any]) -> Union[Categorical, numpy.ndarray]:
    """Sort and deduplicate the values"""
    x = categorized(x)
    if is_categorical_dtype(x):
        lvls = levels(x)
        return factor(lvls, lvls, exclude=NULL, ordered=x.ordered)

    # don't sort on bare list?
    # if isinstance(x, list):
    #     return pandas.unique(x)

    if isinstance(x, DataFrame):
        return arrange(
            distinct(x, __calling_env=CallingEnvs.REGULAR),
            __calling_env=CallingEnvs.REGULAR,
        )

    # return numpy.sort(numpy.unique(x))
    # numpy.unique() will turn ['A', 'B', numpy.nan] to ['A', 'B', 'nan']
    try:
        out = pandas.unique(x)
    except TypeError:
        # unhashable type: 'list'
        # workaround for unhashable elements
        # using its stringified form as key, which has side-effects
        maps = {str(elem): elem for elem in x}
        out = pandas.unique(list(maps.keys()))
        out = numpy.array([maps[elem] for elem in out], dtype=object)

    has_na = is_null(out).any()
    if has_na:
        out = numpy.sort(out[is_not_null(out)])
        return numpy.concatenate([out, [NA]])
    # numpy.sort() cannot do comparisons between string and NA
    return numpy.sort(out)



"""APIs for R-tibble"""
from .tibble import tibble, tibble_row, tribble, zibble
from .verbs import (
    enframe,
    deframe,
    add_row,
    add_case,
    add_column,
    has_rownames,
    has_index,
    remove_index,
    remove_rownames,
    drop_index,
    rownames_to_column,
    index_to_column,
    rowid_to_column,
    column_to_rownames,
    column_to_index,
)



"""Provide DataFrameGroupBy and DataFrameRowwise"""
from typing import Any, Callable, Mapping, Tuple, Union
from abc import ABC, abstractmethod

import numpy
import pandas
from pandas import DataFrame, Series, RangeIndex

from pipda.function import Function, FastEvalFunction
from pipda.symbolic import DirectRefItem, DirectRefAttr, ReferenceAttr
from pipda.utils import CallingEnvs

from .defaults import DEFAULT_COLUMN_PREFIX
from .types import StringOrIter, is_scalar
from .utils import apply_dtypes
from ..base import setdiff


class DataFrameGroupByABC(DataFrame, ABC):
    """Abstract class for DataFrameGroupBy"""

    def __init__(
        self,
        data: Any,
        _group_vars: StringOrIter = None,
        _group_drop: bool = None,
        # used for copy, etc so we don't need to recompute the group data
        _group_data: DataFrame = None,
        **kwargs: Any,
    ) -> None:
        if isinstance(data, DataFrame):
            kwargs["copy"] = True
        super().__init__(data, **kwargs)
        # drop index to align to tidyverse's APIs
        self.reset_index(drop=True, inplace=True)

        # rowwise
        if _group_vars is None:
            _group_vars = []

        if is_scalar(_group_vars):
            _group_vars = [_group_vars]

        # In order to align to dplyr's API
        self.attrs["_group_drop"] = True if _group_drop is None else _group_drop
        self.attrs["_group_vars"] = _group_vars
        self.attrs["_group_data"] = _group_data

    @abstractmethod
    def _datar_apply(
        self,
        _func: Callable,
        *args: Any,
        _mappings: Mapping[str, Any] = None,
        _method: str = "apply",
        _groupdata: bool = True,
        _drop_index: bool = True,
        **kwargs: Any,
    ) -> DataFrame:
        pass

    def copy(
        self,
        deep: bool = True,
        copy_grouped: bool = False,
    ) -> "DataFrameGroupByABC":
        """Copy the dataframe and keep the class"""
        if not copy_grouped:
            return super().copy(deep)

        if deep:
            return self.__class__(
                super().copy(),
                _group_vars=self.attrs["_group_vars"][:],
                _group_drop=self.attrs["_group_drop"],
                _group_data=self._group_data.copy(),
            )
        # we still need to calculate _grouped_df
        return self.__class__(
            self,
            _group_vars=self.attrs["_group_vars"],
            _group_drop=self.attrs["_group_drop"],
            _group_data=self._group_data,
        )


class DataFrameGroupBy(DataFrameGroupByABC):
    """A customized DataFrameGroupBy class, other than pandas' DataFrameGroupBy

    Pandas' DataFrameGroupBy has obj refer to the original data frame. We do it
    the reverse way by attaching the groupby object to the frame. So that it is:
    1. easier to write single dispatch functions, as DataFrameGroupBy is now a
        subclass of pandas' DataFrame
    2. easier to display the frame. We can use all utilities for frame to
        display. By `core._frame_format_patch.py`, we are also able to show the
        grouping information
    3. possible for future optimizations

    Known Issues:
        - Due to https://github.com/pandas-dev/pandas/issues/35202
            Currently `dropna` is fixed to True of `df.groupby(...)`
            So no NAs will be kept in group vars
        - `_drop = FALSE` does not work when there are multiple group vars
        - Since group vars are required in `DataFrame.groupby()`, so virtual
            groupings are not supported.
        - Groupby on a column with tuples creates a multiindex
            https://github.com/pandas-dev/pandas/issues/21340
        - Order of group data/groups does not follow the categories/levels of
            a category group variable.

    Args:
        data: Data that used to construct the frame.
        **kwargs: Additional keyword arguments passed to DataFrame constructor.
        _group_vars: The grouping variables
        _group_drop: Whether to drop non-observable rows
        _group_data: In most cases, used for copy. Use this groupdata
            if provided to avoid recalculate.

    Attributes:
        _grouped_df: The grouped data frame (pandas' DataFrameGroupBy object)
    """

    def __init__(
        self,
        data: Any,
        _group_vars: StringOrIter = None,
        _group_drop: bool = None,
        _group_data: DataFrame = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(data, _group_vars, _group_drop, _group_data, **kwargs)
        self.__dict__["_grouped_df"] = self.groupby(
            _group_vars,
            dropna=True,
            sort=False,
            observed=self.attrs["_group_drop"],
        )

    # @property
    # def _constructor(self):
    #     return DataFrameGroupBy

    @property
    def _group_data(self):
        """The group data"""
        # compose group data using self._grouped_df.grouper
        if self.attrs["_group_data"] is None:
            self.attrs["_group_data"] = DataFrame(
                (
                    [key] + [list(val)]
                    if len(self.attrs["_group_vars"]) == 1
                    else list(key) + [list(val)]
                    for key, val in sorted(
                        self._grouped_df.grouper.groups.items(),
                        key=lambda item: item[1][0]
                        if len(item[1]) > 0
                        else self.shape[0],
                    )
                ),
                columns=self.attrs["_group_vars"] + ["_rows"],
            )
            apply_dtypes(
                self.attrs["_group_data"],
                self.dtypes[self.attrs["_group_vars"]].to_dict(),
            )

        return self.attrs["_group_data"]

    def _datar_apply(
        self,
        _func: Callable,
        *args: Any,
        _mappings: Mapping[str, Any] = None,
        _method: str = "apply",
        _groupdata: bool = True,
        _drop_index: bool = True,
        **kwargs: Any,
    ) -> DataFrame:
        """Customized apply.

        Aggregation on single columns will be tried to optimized.

        Args:
            _func: The function to be applied.
            *args: The non-keyword arguments for the function
            _mappings: The mapping (new name => transformation/aggregation)
                This will be used to check if we can turn the apply into agg.
                transformations like `f.x.mean()` or `mean(f.x)` will use agg
                to do the aggregation.
                `_func` will be used to apply if any item fails to be optimized.
            _method: Only optimize single column transformations
                when `agg` is provided
            _groupdata: Whether attach the group data to the result data frame
                anyway. Pandas will lose them if is a transform (num. of
                rows of result data frame is equal to the source data frame)
                instead of an aggregation
            _drop_index: Whether we should drop the index when the
                sub data frame goes into the apply function (`_func`).

        Returns:
            The transformed/aggregated data frame
        """
        optms = _optimizable(_mappings)
        if _method == "agg" and optms:
            out = self._datar_agg(optms)

        else:
            group_index = -1

            def _applied(subdf):
                nonlocal group_index
                group_index += 1
                if _drop_index:
                    subdf = subdf.reset_index(drop=True)
                subdf.attrs["_group_index"] = group_index
                subdf.attrs["_group_data"] = self._group_data
                ret = _func(subdf, *args, **kwargs)
                return None if ret is None else ret

            # keep the order
            out = self._grouped_df.apply(_applied).sort_index(level=-1)

        if not _groupdata:
            return out.reset_index(drop=True)

        if (
            self.shape[0] > 0
            and self.shape[0] == out.shape[0]
            and out.index.names == [None]
        ):
            gkeys = self._group_data[
                self._group_data.columns[:-1].difference(out.columns)
            ]
            return pandas.concat([gkeys, out], axis=1)

        # in case the group vars are mutated
        index_to_reset = setdiff(
            self.attrs["_group_vars"],
            out.columns,
            __calling_env=CallingEnvs.REGULAR,
        )
        return out.reset_index(level=index_to_reset).reset_index(drop=True)

    def _datar_agg(
        self,
        _mappings: Mapping[str, Any],
    ) -> DataFrame:
        """Apply aggregation to the mappings"""
        return self._grouped_df.agg(**_mappings)


class DataFrameRowwise(DataFrameGroupBy):
    """Rowwise dataframe"""

    __init__ = DataFrameGroupByABC.__init__

    @property
    def _group_data(self):
        """The group data with the ori"""
        if self.attrs["_group_data"] is None:
            self.attrs["_group_data"] = self[self.attrs["_group_vars"]].assign(
                _rows=[[i] for i in range(self.shape[0])]
            )

        return self.attrs["_group_data"]

    def _datar_apply(
        self,
        _func: Callable,
        *args: Any,
        _mappings: Mapping[str, Any] = None,
        _method: str = "apply",
        _groupdata: bool = True,
        _drop_index: bool = True,
        **kwargs: Any,
    ) -> DataFrame:
        # TODO: check if _func is optimizable.
        def applied(ser, *args, **kwargs):
            """Make sure a series is returned"""
            ret = _func(ser, *args, **kwargs)
            if isinstance(ret, DataFrame) and ret.shape[0] > 1:
                raise ValueError(
                    "Must return a 1d array, a Series or a DataFrame "
                    "with 1 row when a function is applied to "
                    "a DataFrameRowwise object."
                )

            if isinstance(ret, DataFrame):
                ret = ret.iloc[0, :]

            if ret is None:
                ret = []

            if is_scalar(ret):
                ret = [ret]

            if len(ret) == 0:
                ret = Series(ret, dtype=getattr(ret, "dtype", object))
            else:
                ret = Series(ret)

            if isinstance(ret.index, RangeIndex):
                ret.index = (
                    f"{DEFAULT_COLUMN_PREFIX}{i}" for i in range(len(ret))
                )

            return ret

        out = self.apply(
            applied,
            axis=1,
            raw=False,
            # result_type='reduce',
            args=args,
            **kwargs,
        )
        if not _groupdata or not self.attrs["_group_vars"]:
            return out

        return pandas.concat([self[self.attrs["_group_vars"]], out], axis=1)


def _optimizable_func(
    name: str, kwargs: Mapping[str, Any]
) -> Union[str, Callable]:
    """Get the name of the function that can be optimized.

    Currently, only functions avaiable with numpy is able to be optimized

    Args:
        name: The name of the function
        kwargs: The keyword arguments to be passed to the function
            Current, if `na_rm` is `True`, then `nan` version of the function
            will be used

    Return:
        The name of the function that can be optimized.
    """
    if not hasattr(numpy, name):
        return None

    if kwargs.get("na_rm", False):
        # Only optimize when na_rm is True, because
        # pandas will ignore NAs anyway
        return getattr(numpy, f"nan{name}")

    # return getattr(numpy, name)
    # refuse to optimize, as NAs get lost by pandas
    return None


def _optimizable(
    mappings: Mapping[str, Any]
) -> Mapping[str, Tuple[str, Union[str, Callable]]]:
    """Turn mappings (new column name => transform/aggregation) to
    (new column name => tuple(old column name, aggregation function))
    """
    if mappings is None:
        return None

    out = {}
    for key, val in mappings.items():
        # optimize constants?
        if (
            isinstance(val, FastEvalFunction)
            and len(val._pipda_args) == 1
            and isinstance(val._pipda_args[0], (DirectRefItem, DirectRefAttr))
        ):
            # mean(f.x)
            ofun = _optimizable_func(
                val._pipda_func.__name__, val._pipda_kwargs
            )
            if not ofun:
                return None
            out[key] = (val._pipda_args[0]._pipda_ref, ofun)
        elif (
            isinstance(val, Function)
            and isinstance(val._pipda_func, ReferenceAttr)
            and isinstance(
                val._pipda_func._pipda_parent, (DirectRefItem, DirectRefAttr)
            )
        ):
            # f.x.mean()
            out[key] = (
                val._pipda_func._pipda_parent._pipda_ref,
                val._pipda_func._pipda_ref,
            )
        else:
            return None
    return out



"""Factors, implemented with pandas' Categorical

The huge difference:
R's factors support NAs in levels but Categorical cannot have NAs in categories.
"""

from typing import Any, Iterable

from pandas import Categorical
from pipda import register_func

from ..core.contexts import Context
from ..core.types import (
    is_scalar,
    is_categorical as is_categorical_,
    ArrayLikeType,
)
from ..core.utils import categorized

from .na import NA


@register_func(None, context=Context.EVAL)
def droplevels(x: Categorical) -> Categorical:
    """drop unused levels from a factor

    Args:
        x: The categorical data

    Returns:
        The categorical data with unused categories dropped.
    """
    return categorized(x).remove_unused_categories()


@register_func(None, context=Context.EVAL)
def levels(x: Any) -> ArrayLikeType:
    """Get levels from a factor

    Args:
        x: The categorical data

    Returns:
        levels of the categorical
        None if x is not an categorical/factor
    """
    if not is_categorical_(x):
        return None

    return categorized(x).categories.values.copy()


@register_func(None, context=Context.EVAL)
def nlevels(x: Any) -> int:
    """Get the number of levels of a factor

    Args:
        x: The data to get number of levels of

    Returns:
        Number of levels if x is a categorical/factor; otherwise 0
    """
    lvls = levels(x)
    return 0 if lvls is None else len(lvls)


@register_func(None, context=Context.EVAL)
def is_ordered(x: Any) -> bool:
    """Check if a factor is ordered"""
    if not is_categorical_(x):
        return False

    return categorized(x).ordered


def factor(
    x: Iterable[Any] = None,

    levels: Iterable[Any] = None,
    exclude: Any = NA,
    ordered: bool = False,
) -> Categorical:
    """encode a vector as a factor (the terms ‘category’ and ‘enumerated type’
    are also used for factors).

    If argument ordered is TRUE, the factor levels are assumed to be ordered

    Args:
        x: a vector of data
        levels: an optional vector of the unique values (as character strings)
            that x might have taken.
        exclude: a vector of values to be excluded when forming the set of
            levels. This may be factor with the same level set as x or
            should be a character
        ordered: logical flag to determine if the levels should be regarded
            as ordered (in the order given).
    """
    if x is None:
        x = []

    # pandas v1.3.0
    # FutureWarning: Allowing scalars in the Categorical constructor
    # is deprecated and will raise in a future version.
    if is_scalar(x):
        x = [x]

    if is_categorical_(x):
        x = x.to_numpy()

    ret = Categorical(x, categories=levels, ordered=ordered)
    if exclude in [False, None]:
        return ret

    if is_scalar(exclude):
        exclude = [exclude]

    return ret.remove_categories(exclude)


def ordered(
    x: Iterable[Any] = None,

    levels: Iterable[Any] = None,
) -> Categorical:
    """Create an ordered factor

    Args:
        x: The values to create factor

    Returns:
        The ordered factor
    """
    return factor(x, levels=levels).as_ordered()


@register_func(None, context=Context.EVAL)
def as_factor(x: Iterable) -> Categorical:
    """Convert an iterable into a pandas.Categorical object

    Args:
        x: The iterable

    Returns:
        The converted categorical object
    """
    return Categorical(x)


as_categorical = as_factor


@register_func(None, context=Context.EVAL)
def is_categorical(x: Any) -> bool:
    """Check if x is categorical data

    Alias `is_factor`

    Args:
        x: The value to be checked

    Returns:
        True if `x` is categorical else False
    """
    return is_categorical_(x)


is_factor = is_categorical



"""Functions ported from tidyverse-tibble"""
from typing import Any, Callable, Iterable, List, Mapping, Union

from pandas import DataFrame, Series, RangeIndex
from pipda import Context, register_verb
from pipda.utils import CallingEnvs

from ..core.utils import (
    copy_attrs,
    get_option,
    position_after,
    position_at,
    recycle_value,
    logger,
    reconstruct_tibble,
)
from ..core.grouped import DataFrameGroupBy, DataFrameRowwise
from ..core.types import is_scalar, Dtype
from ..core.exceptions import ColumnNotExistingError
from ..base import setdiff

from .tibble import tibble


def enframe(
    x: Union[Iterable, Mapping],
    name: str = "name",
    value: str = "value",
    base0_: bool = None,
) -> DataFrame:
    """Converts mappings or lists to one- or two-column data frames.

    Args:
        x: a list, a dictionary or a dataframe with one or two columns
        name: and
        value: value Names of the columns that store the names and values.
            If `None`, a one-column dataframe is returned.
            `value` cannot be `None`
        base0_: Whether the indexes for lists converted to name are 0-based
            or not.

    Returns:
        A data frame with two columns if `name` is not None (default) or
        one-column otherwise.
    """
    if not value:
        raise ValueError("`value` can't be empty.")

    if x is None:
        x = []
    if is_scalar(x):
        x = [x]

    if len(getattr(x, "shape", ())) > 1:
        raise ValueError(
            f"`x` must not have more than one dimension, got {len(x.shape)}."
        )

    if not name and isinstance(x, dict):
        x = x.values()

    elif name:
        if not isinstance(x, dict):
            base0_ = get_option("index.base.0", base0_)
            names = (i + int(not base0_) for i in range(len(x)))
            values = x
        else:
            names = x.keys()
            values = x.values()
        x = (list(item) for item in zip(names, values))

    return DataFrame(x, columns=[name, value] if name else [value])


def deframe(x: DataFrame) -> Union[Iterable, Mapping]:
    """Converts two-column data frames to a dictionary
    using the first column as name and the second column as value.
    If the input has only one column, a list.

    Args:
        x: A data frame.

    Returns:
        A dictionary or a list if only one column in the data frame.
    """
    if x.shape[1] == 1:
        return x.iloc[:, 0].values

    if x.shape[1] != 2:
        logger.warning(
            "`x` must be a one- or two-column data frame in `deframe()`."
        )

    return dict(zip(x.iloc[:, 0], x.iloc[:, 1]))


@register_verb(
    DataFrame,
    context=Context.EVAL,
    extra_contexts={"_before": Context.SELECT, "_after": Context.SELECT},
)
def add_row(
    _data: DataFrame,
    *args: Any,
    _before: int = None,
    _after: int = None,
    base0_: bool = None,
    **kwargs: Any,
) -> DataFrame:
    """Add one or more rows of data to an existing data frame.

    Aliases `add_case`

    Args:
        _data: Data frame to append to.
        *args: and
        **kwargs: Name-value pairs to add to the data frame.
        _before: and
        _after: row index where to add the new rows.
            (default to add after the last row)
        base0_: Whether `_before` and `_after` are 0-based or not.

    Returns:
        The dataframe with the added rows

    """
    if isinstance(_data, DataFrameGroupBy) and not isinstance(
        _data, DataFrameRowwise
    ):
        raise ValueError("Can't add rows to grouped data frames.")

    if not args and not kwargs:
        df = DataFrame(index=[0], columns=_data.columns)
    else:
        df = tibble(*args, **kwargs, frame_=2)
        if df.shape[0] == 0:
            for col in _data.columns:
                df[col] = Series(dtype=_data[col].dtype)

    extra_vars = setdiff(
        df.columns,
        _data.columns,
        __calling_env=CallingEnvs.REGULAR,
    )
    if extra_vars:
        raise ValueError(f"New rows can't add columns: {extra_vars}")

    pos = _pos_from_before_after(_before, _after, _data.shape[0], base0_)
    out = _rbind_at(_data, df, pos)

    if isinstance(_data, DataFrameRowwise):
        out = reconstruct_tibble(_data, out, keep_rowwise=True)
    else:
        copy_attrs(out, _data)
    return out


add_case = add_row


@register_verb(
    DataFrame,
    context=Context.EVAL,
    extra_contexts={"_before": Context.SELECT, "_after": Context.SELECT},
)
def add_column(
    _data: DataFrame,
    *args: Any,
    _before: Union[str, int] = None,
    _after: Union[str, int] = None,
    _name_repair: Union[str, Callable] = "check_unique",
    base0_: bool = None,
    dtypes_: Union[Dtype, Mapping[str, Dtype]] = None,
    **kwargs: Any,
) -> DataFrame:
    """Add one or more columns to an existing data frame.

    Args:
        _data: Data frame to append to
        *args: and
        **kwargs: Name-value pairs to add to the data frame
        _before: and
        _after: Column index or name where to add the new columns
            (default to add after the last column)
        base0_: Whether `_before` and `_after` are 0-based if they are index.
            if not given, will be determined by `get_option('index_base_0')`,
            which is `False` by default.
        dtypes_: The dtypes for the new columns, either a uniform dtype or a
            dict of dtypes with keys the column names

    Returns:
        The dataframe with the added columns
    """
    df = tibble(
        *args, **kwargs, _name_repair="minimal", dtypes_=dtypes_, frame_=2
    )

    if df.shape[1] == 0:
        return _data.copy()

    df = recycle_value(df, len(_data), "new columns")
    pos = _pos_from_before_after_names(
        _before, _after, _data.columns.tolist(), base0_
    )

    out = _cbind_at(_data, df, pos, _name_repair)
    if len(_data) == 0:
        out = out.loc[[], :]
    return reconstruct_tibble(_data, out, keep_rowwise=True)


@register_verb(DataFrame)
def has_rownames(_data: DataFrame) -> bool:
    """Detect if a data frame has row names

    Aliases `has_index`

    Args:
        _data: The data frame to check

    Returns:
        True if the data frame has index otherwise False.

    """
    return not isinstance(_data.index, RangeIndex)


has_index = has_rownames


@register_verb(DataFrame)
def remove_rownames(_data: DataFrame) -> DataFrame:
    """Remove the index/rownames of a data frame

    Aliases `remove_index`, `drop_index`, `remove_rownames`

    Args:
        _data: The data frame

    Returns:
        The data frame with index removed

    """
    return _data.reset_index(drop=True)


remove_index = drop_index = remove_rownames


@register_verb(DataFrame, context=Context.SELECT)
def rownames_to_column(_data: DataFrame, var="rowname") -> DataFrame:
    """Add rownames as a column

    Aliases `index_to_column`

    Args:
        _data: The data frame
        var: The name of the column

    Returns:
        The data frame with rownames added as one column. Note that the
        original index is removed.
    """
    if var in _data.columns:
        raise ValueError(f"Column name `{var}` must not be duplicated.")

    from ..dplyr.mutate import mutate

    return remove_rownames(
        mutate(
            _data,
            **{var: _data.index},
            _before=1,
            __calling_env=CallingEnvs.REGULAR,
        ),
        __calling_env=CallingEnvs.REGULAR,
    )


index_to_column = rownames_to_column


@register_verb(DataFrame, context=Context.SELECT)
def rowid_to_column(
    _data: DataFrame, var="rowid", base0_: bool = False
) -> DataFrame:
    """Add rownames as a column

    Args:
        _data: The data frame
        var: The name of the column

    Returns:
        The data frame with row ids added as one column.

    """
    if var in _data.columns:
        raise ValueError(f"Column name `{var}` must not be duplicated.")

    from ..dplyr.mutate import mutate

    base = int(not base0_)
    return remove_rownames(
        mutate(
            _data,
            **{var: range(base, _data.shape[0] + base)},
            _before=1,
            __calling_env=CallingEnvs.REGULAR,
        ),
        __calling_env=CallingEnvs.REGULAR,
    )


@register_verb(DataFrame, context=Context.SELECT)
def column_to_rownames(_data: DataFrame, var: str = "rowname") -> DataFrame:
    """Set rownames/index with one column, and remove it

    Aliases `column_to_index`

    Args:
        _data: The data frame
        var: The column to conver to the rownames

    Returns:
        The data frame with the column converted to rownames
    """
    if has_rownames(_data, __calling_env=CallingEnvs.REGULAR):
        raise ValueError("`_data` must be a data frame without row names.")

    from ..dplyr.mutate import mutate

    try:
        rownames = [str(name) for name in _data[var]]
    except KeyError:
        raise ColumnNotExistingError(
            f"Column `{var}` does not exist."
        ) from None
    out = mutate(_data, **{var: None}, __calling_env=CallingEnvs.REGULAR)
    out.index = rownames
    return out


column_to_index = column_to_rownames

# Helpers ------------------------------------------------------------------


def _pos_from_before_after_names(
    before: Union[str, int],
    after: Union[str, int],
    names: List[str],
    base0: bool,
) -> int:
    """Get the position to insert from before and after"""
    if before is not None:
        before, base0 = _check_names_before_after(before, names, base0)
    if after is not None:
        after, base0 = _check_names_before_after(after, names, base0)
    return _pos_from_before_after(before, after, len(names), base0)


def _check_names_before_after(
    pos: Union[str, int], names: List[str], base0: bool
) -> int:
    """Get position by given index or name"""
    if not isinstance(pos, str):
        return pos, base0
    try:
        return names.index(pos), True
    except ValueError:
        raise ColumnNotExistingError(
            f"Column `{pos}` does not exist."
        ) from None


def _cbind_at(
    data: DataFrame, df: DataFrame, pos: int, _name_repair: Union[str, Callable]
) -> DataFrame:
    """Column bind at certain pos, 0-based"""
    from ..dplyr import bind_cols

    part1 = data.iloc[:, :pos]
    part2 = data.iloc[:, pos:]
    return bind_cols(
        part1,
        df,
        part2,
        _name_repair=_name_repair,
        __calling_env=CallingEnvs.REGULAR,
    )


def _pos_from_before_after(
    before: int, after: int, length: int, base0: bool
) -> int:
    """Get the position to insert from before and after"""
    if before is not None and after is not None:
        raise ValueError("Can't specify both `_before` and `_after`.")

    if before is None and after is None:
        return length

    if after is not None:
        return position_after(after, length, base0)
    return position_at(before, length, base0)


def _rbind_at(data: DataFrame, df: DataFrame, pos: int) -> DataFrame:
    """Row bind at certain pos, 0-based"""
    from ..dplyr import bind_rows

    part1 = data.iloc[:pos, :]
    part2 = data.iloc[pos:, :]
    return bind_rows(part1, df, part2, __calling_env=CallingEnvs.REGULAR)



"""Exceptions for datar"""


class DatarException(Exception):
    """Base exectpion for datar"""


class ColumnNotExistingError(DatarException):
    """When selecting non-existing columns"""


class ColumnNameInvalidError(DatarException):
    """When invalid column names provided to select"""


class DataUnalignableError(DatarException):
    """When two data cannot be aligned to each other"""


class NameNonUniqueError(DatarException):
    """When check_unique fails"""


class DataUnrecyclable(DatarException):
    """When data is not recyclable"""



"""Lower-level APIs to manipulate the factors"""
from typing import Iterable, List
from pandas import Categorical
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.types import IntType, ForcatsRegType, ForcatsType
from ..core.utils import position_at
from ..core.contexts import Context
from ..base import setdiff, union

from ..base import (
    NA,
    as_character,
    as_integer,
    setequal,
    seq_along,
    factor,
    levels,
    match,
    nlevels,
    is_ordered,
    is_integer,
    unique,
)

from ..dplyr import recode_factor

from .utils import check_factor


def lvls_seq(_f: ForcatsType, base0_: bool = None):
    """Get the index sequence of a factor levels"""
    return seq_along(
        levels(_f, __calling_env=CallingEnvs.REGULAR),
        base0_=base0_,
        __calling_env=CallingEnvs.REGULAR,
    )


def refactor(_f, new_levels: Iterable, ordered: bool = None) -> Categorical:
    """Refactor using new levels"""
    if ordered is None:
        ordered = is_ordered(_f, __calling_env=CallingEnvs.REGULAR)

    new_f = factor(_f, levels=new_levels, exclude=NA, ordered=ordered)
    # keep attributes?
    return new_f


@register_verb(ForcatsRegType, context=Context.EVAL)
def lvls_reorder(
    _f: ForcatsType,
    idx: Iterable[IntType],
    ordered: bool = None,
    base0_: bool = None,
) -> Categorical:
    """Leaves values of a factor as they are, but changes the order by
    given indices

    Args:
        f: A factor (or character vector).
        idx: A integer index, with one integer for each existing level.
        new_levels: A character vector of new levels.
        ordered: A logical which determines the "ordered" status of the
          output factor. `None` preserves the existing status of the factor.

    Returns:
        The factor with levels reordered
    """
    _f = check_factor(_f)
    if not is_integer(idx, __calling_env=CallingEnvs.REGULAR):
        raise ValueError("`idx` must be integers")

    len_idx = len(idx)
    seq_lvls = lvls_seq(_f, base0_=base0_)
    if not setequal(
        idx, seq_lvls, __calling_env=CallingEnvs.REGULAR
    ) or len_idx != nlevels(_f, __calling_env=CallingEnvs.REGULAR):
        raise ValueError("`idx` must contain one integer for each level of `f`")

    pos = [position_at(index, len_idx, base0=base0_) for index in idx]
    return refactor(
        _f,
        levels(_f, __calling_env=CallingEnvs.REGULAR)[pos],
        ordered=ordered,
    )


@register_verb(ForcatsRegType, context=Context.EVAL)
def lvls_revalue(
    _f: ForcatsType,
    new_levels: Iterable,
) -> Categorical:
    """changes the values of existing levels; there must
    be one new level for each old level

    Args:
        _f: A factor
        new_levels: A character vector of new levels.

    Returns:
        The factor with the new levels
    """
    _f = check_factor(_f)

    if len(new_levels) != nlevels(_f, __calling_env=CallingEnvs.REGULAR):
        raise ValueError(
            "`new_levels` must be the same length as `levels(f)`: expected ",
            f"{nlevels(_f, __calling_env=CallingEnvs.REGULAR)} new levels, "
            f"got {len(new_levels)}.",
        )

    u_levels = unique(new_levels, __calling_env=CallingEnvs.REGULAR)
    if len(new_levels) > len(u_levels):
        # has duplicates
        index = match(new_levels, u_levels, __calling_env=CallingEnvs.REGULAR)
        out = factor(
            as_character(
                index[
                    as_integer(
                        _f,
                        base0_=True,
                        __calling_env=CallingEnvs.REGULAR,
                    )
                ],
                __calling_env=CallingEnvs.REGULAR,
            )
        )
        return recode_factor(
            out,
            dict(
                zip(
                    levels(out, __calling_env=CallingEnvs.REGULAR),
                    u_levels,
                )
            ),
            __calling_env=CallingEnvs.REGULAR,
        )

    recodings = dict(
        zip(levels(_f, __calling_env=CallingEnvs.REGULAR), new_levels)
    )
    return recode_factor(_f, recodings, __calling_env=CallingEnvs.REGULAR)


@register_verb(ForcatsRegType, context=Context.EVAL)
def lvls_expand(
    _f: ForcatsType,
    new_levels: Iterable,
) -> Categorical:
    """Expands the set of levels; the new levels must
    include the old levels.

    Args:
        _f: A factor
        new_levels: The new levels. Must include the old ones

    Returns:
        The factor with the new levels
    """
    _f = check_factor(_f)
    levs = levels(_f, __calling_env=CallingEnvs.REGULAR)

    missing = setdiff(levs, new_levels, __calling_env=CallingEnvs.REGULAR)
    if len(missing) > 0:
        raise ValueError("Must include all existing levels. Missing: {missing}")

    return refactor(_f, new_levels=new_levels)


@register_verb(ForcatsRegType)
def lvls_union(
    fs: Iterable[ForcatsType],
) -> List:
    """Find all levels in a list of factors

    Args:
        fs: A list of factors

    Returns:
        A list of all levels
    """
    out = []
    for fct in fs:
        fct = check_factor(fct)
        levs = levels(fct, __calling_env=CallingEnvs.REGULAR)
        out = union(out, levs, __calling_env=CallingEnvs.REGULAR)
    return out



"""Context dependent expressions

See souce https://github.com/tidyverse/dplyr/blob/master/R/context.R
"""
from typing import Any, Iterable, List

from pandas import DataFrame
from pipda import register_func
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.middlewares import CurColumn
from ..base import setdiff
from .group_data import group_vars


# n used directly in count
@register_func(context=Context.EVAL, summarise_prefers_input=True)
def n(series: Iterable[Any]) -> int:
    """gives the current group size."""
    return len(series)


@register_func(DataFrame, verb_arg_only=True, summarise_prefers_input=True)
def cur_data_all(_data: DataFrame) -> DataFrame:
    """gives the current data for the current group
    (including grouping variables)"""
    return _data


@register_func(DataFrame, verb_arg_only=True, summarise_prefers_input=True)
def cur_data(_data: DataFrame) -> int:
    """gives the current data for the current group
    (excluding grouping variables)."""
    return _data[
        setdiff(
            _data.columns,
            group_vars(_data, __calling_env=CallingEnvs.REGULAR),
            __calling_env=CallingEnvs.REGULAR,
        )
    ]


@register_func(DataFrame, verb_arg_only=True, summarise_prefers_input=True)
def cur_group(_data: DataFrame) -> DataFrame:
    """gives the group keys, a tibble with one row and one column for
    each grouping variable."""
    index = _data.attrs.get("_group_index", None)
    if index is None:
        return DataFrame(index=range(_data.shape[0]))

    gdata = _data.attrs["_group_data"]
    return gdata.iloc[[index], :-1]


@register_func(DataFrame, verb_arg_only=True, summarise_prefers_input=True)
def cur_group_id(_data: DataFrame) -> int:
    """gives a unique numeric identifier for the current group."""
    return _data.attrs.get("_group_index", 1)


@register_func(DataFrame, verb_arg_only=True, summarise_prefers_input=True)
def cur_group_rows(_data: DataFrame) -> List[int]:
    """Gives the row indices for the current group.

    Args:
        _data: The dataFrame.

    Returns:
        The `_rows` from group data or row indexes (always 0-based).
    """
    index = _data.attrs.get("_group_index", None)
    if index is None:
        return list(range(_data.shape[0]))
    return _data.attrs["_group_data"].loc[index, "_rows"]


def cur_column() -> CurColumn:
    """Used in the functions of across. So we don't have to register it."""
    return CurColumn()



"""Create, modify, and delete columns

See source https://github.com/tidyverse/dplyr/blob/master/R/mutate.R
"""

from typing import Any, Tuple, List, Union
from pandas import DataFrame, Series
from pipda import register_verb, evaluate_expr, ContextBase
from pipda.utils import CallingEnvs

from ..core.contexts import Context, ContextEval
from ..core.utils import (
    dedup_name,
    recycle_value,
    arg_match,
    df_setitem,
    name_mutatable_args,
    reconstruct_tibble,
)
from ..core.defaults import DEFAULT_COLUMN_PREFIX
from ..core.grouped import DataFrameGroupBy
from ..base import setdiff, union, intersect, c, NA
from .group_by import group_by_drop_default
from .group_data import group_vars, group_data
from .relocate import relocate


@register_verb(
    DataFrame,
    context=Context.PENDING,
    extra_contexts={"_before": Context.SELECT, "_after": Context.SELECT},
)
def mutate(
    _data: DataFrame,
    *args: Any,
    _keep: str = "all",
    _before: Union[int, str] = None,
    _after: Union[int, str] = None,
    base0_: bool = None,
    **kwargs: Any,
) -> DataFrame:

    """Adds new variables and preserves existing ones

    The original API:
    https://dplyr.tidyverse.org/reference/mutate.html

    Args:
        _data: A data frame
        _keep: allows you to control which columns from _data are retained
            in the output:
            - "all", the default, retains all variables.
            - "used" keeps any variables used to make new variables;
              it's useful for checking your work as it displays inputs and
              outputs side-by-side.
            - "unused" keeps only existing variables not used to make new
                variables.
            - "none", only keeps grouping keys (like transmute()).
        _before: and
        _after: Optionally, control where new columns should appear
            (the default is to add to the right hand side).
            See relocate() for more details.
        base0_: Whether `_before` and `_after` are 0-based if given by indexes.
            If not provided, will use `datar.base.get_option('index.base.0')`
        *args: and
        **kwargs: Name-value pairs. The name gives the name of the column
            in the output. The value can be:
            - A vector of length 1, which will be recycled to the correct
                length.
            - A vector the same length as the current group (or the whole
                data frame if ungrouped).
            - None to remove the column

    Returns:
        An object of the same type as _data. The output has the following
        properties:
        - Rows are not affected.
        - Existing columns will be preserved according to the _keep
            argument. New columns will be placed according to the
            _before and _after arguments. If _keep = "none"
            (as in transmute()), the output order is determined only
            by ..., not the order of existing columns.
        - Columns given value None will be removed
        - Groups will be recomputed if a grouping variable is mutated.
        - Data frame attributes are preserved.
    """
    keep = arg_match(_keep, "_keep", ["all", "unused", "used", "none"])

    context = ContextEval()
    cols, removed = _mutate_cols(_data, context, *args, **kwargs)
    if cols is None:
        cols = DataFrame(index=_data.index)

    out = _data.copy()
    # order is the same as _data
    out[cols.columns.tolist()] = cols
    # out.columns.difference(removed)
    # changes column order when removed == []
    out = out[setdiff(out.columns, removed, __calling_env=CallingEnvs.REGULAR)]
    if _before is not None or _after is not None:
        new = setdiff(
            cols.columns,
            _data.columns,
            __calling_env=CallingEnvs.REGULAR,
        )
        out = relocate(
            out,
            *new,
            _before=_before,
            _after=_after,
            base0_=base0_,
            __calling_env=CallingEnvs.REGULAR,
        )

    if keep == "all":
        keep = out.columns
    elif keep == "unused":
        used = context.used_refs.keys()
        unused = setdiff(
            _data.columns,
            used,
            __calling_env=CallingEnvs.REGULAR,
        )
        keep = intersect(
            out.columns,
            c(
                group_vars(_data, __calling_env=CallingEnvs.REGULAR),
                unused,
                cols.columns,
            ),
            __calling_env=CallingEnvs.REGULAR,
        )
    elif keep == "used":
        used = context.used_refs.keys()
        keep = intersect(
            out.columns,
            c(
                group_vars(_data, __calling_env=CallingEnvs.REGULAR),
                used,
                cols.columns,
            ),
            __calling_env=CallingEnvs.REGULAR,
        )
    else:  # keep == 'none':
        keep = union(
            setdiff(
                group_vars(_data, __calling_env=CallingEnvs.REGULAR),
                cols.columns,
                __calling_env=CallingEnvs.REGULAR,
            ),
            intersect(
                cols.columns,
                out.columns,
                __calling_env=CallingEnvs.REGULAR,
            ),
            __calling_env=CallingEnvs.REGULAR,
        )

    out = out[keep]
    return out.loc[[], :] if len(_data) == 0 else out


@mutate.register(DataFrameGroupBy, context=Context.PENDING)
def _(
    _data: DataFrameGroupBy,
    *args: Any,
    _keep: str = "all",
    _before: str = None,
    _after: str = None,
    base0_: bool = None,
    **kwargs: Any,
) -> DataFrameGroupBy:
    """Mutate on DataFrameGroupBy object"""

    def apply_func(df):
        if isinstance(df, Series):
            df = df.to_frame().T
            index = df.attrs["_group_index"] = df.index[0]
            df.attrs["_group_data"] = _data._group_data
            rows = [index]
        else:
            index = df.attrs["_group_index"]
            rows = df.attrs["_group_data"].loc[index, "_rows"]

        ret = mutate(
            df.reset_index(drop=True),
            *args,
            _keep=_keep,
            _before=_before,
            _after=_after,
            base0_=base0_,
            __calling_env=CallingEnvs.REGULAR,
            **kwargs,
        )
        ret.index = rows
        return ret

    out = _data._datar_apply(apply_func, _drop_index=False).sort_index()
    if out.shape[0] > 0:
        # keep the original row order
        # out.sort_index(inplace=True)
        # not only DataFrameGroupBy but also DataFrameRowwise
        return reconstruct_tibble(_data, out, keep_rowwise=True)

    # 0-row
    named = name_mutatable_args(*args, **kwargs)
    df = DataFrame({key: [] for key in named})
    out = _data.copy()
    out[df.columns.tolist()] = df
    return _data.__class__(
        out,
        _group_vars=group_vars(_data, __calling_env=CallingEnvs.REGULAR),
        _group_drop=group_by_drop_default(_data),
        _group_data=group_data(_data, __calling_env=CallingEnvs.REGULAR),
    )


@register_verb(DataFrame, context=Context.PENDING)
def transmute(
    _data: DataFrame,
    *args: Any,
    _before: Union[int, str] = None,
    _after: Union[int, str] = None,
    base0_: bool = None,
    **kwargs: Any,
) -> DataFrame:
    """Mutate with _keep='none'

    See Also:
        [`mutate()`](datar.dplyr.mutate.mutate).
    """
    return mutate(
        _data,
        *args,
        _keep="none",
        _before=_before,
        _after=_after,
        base0_=base0_,
        __calling_env=CallingEnvs.REGULAR,
        **kwargs,
    )


def _mutate_cols(
    _data: DataFrame,
    _context: ContextBase,
    *args: Any,
    **kwargs: Any,
) -> Tuple[DataFrame, List[str]]:
    """Mutate columns"""
    if not args and not kwargs:
        return None, []
    _data = _data.copy()
    named_mutatables = name_mutatable_args(*args, **kwargs)
    new_columns = []
    removed = []
    add_new_name = True
    for name, mutatable in named_mutatables.items():
        ddp_name = dedup_name(name, list(named_mutatables))
        # if not a dedup name, it's a new name
        add_new_name = ddp_name == name
        mutatable = evaluate_expr(mutatable, _data, _context)
        if mutatable is None:
            if ddp_name in _data:
                removed.append(ddp_name)
                _data.drop(columns=[ddp_name], inplace=True)
            # be silent if name doesn't exist
            continue

        if isinstance(mutatable, DataFrame):
            if mutatable.shape[1] == 0 and not ddp_name.startswith(
                DEFAULT_COLUMN_PREFIX
            ):
                _data = df_setitem(
                    _data, ddp_name, [NA] * max(mutatable.shape[0], 1)
                )
                if add_new_name:
                    new_columns.append(ddp_name)
            else:
                for col in mutatable.columns:
                    new_name = (
                        col
                        if ddp_name.startswith(DEFAULT_COLUMN_PREFIX)
                        else f"{ddp_name}${col}"
                    )
                    coldata = recycle_value(
                        mutatable[col], _data.shape[0], ddp_name
                    )
                    _data = df_setitem(_data, new_name, coldata)

                    if add_new_name:
                        new_columns.append(new_name)
        else:
            mutatable = recycle_value(mutatable, _data.shape[0], ddp_name)
            _data = df_setitem(_data, ddp_name, mutatable)

            if add_new_name:
                new_columns.append(ddp_name)

    # keep column order
    return _data[new_columns], removed



"""Functions from R-dplyr"""
from typing import Any, Iterable

import numpy
from pandas import DataFrame, Series
from pipda import register_func

from ..core.types import (
    BoolOrIter,
    NumericOrIter,
    NumericType,
    is_iterable,
    is_scalar,
    is_null,
)
from ..core.contexts import Context
from ..core.utils import position_at, Array
from ..base import NA


@register_func(None, context=Context.EVAL)
def between(
    x: NumericOrIter, left: NumericType, right: NumericType
) -> BoolOrIter:
    """Function version of `left <= x <= right`, works for both scalar and
    vector data

    See https://dplyr.tidyverse.org/reference/between.html

    Args:
        x: The data to test
        left: and
        right: The boundary values (must be scalars)

    Returns:
        A bool value of `x` is scalar, otherwise a `Series` of boolean values
    """
    if not is_scalar(left) or not is_scalar(right):
        raise ValueError(f"`{between}` expects scalars for `left` and `right`.")
    if is_scalar(x):
        if is_null([x, left, right]).any():
            return NA
        return left <= x <= right
    return Series(between(elem, left, right) for elem in x)


@register_func(None, context=Context.EVAL)
def cummean(series: Iterable[NumericType]) -> Iterable[float]:
    """Get cumulative means"""
    if not isinstance(series, Series):
        series = Series(series, dtype="float64")
    if len(series) > 0:
        return series.cumsum(skipna=False) / (Series(range(len(series))) + 1.0)
    return Series([], dtype="float64")


@register_func(None, context=Context.EVAL)
def cumall(series: Any) -> Series:
    """Get cumulative bool. All cases after first False"""
    if is_scalar(series):
        series = [series]

    bools = boolean(series)  # all to bool, with NAs kept
    out = []
    out_append = out.append
    for elem in bools:
        if not out:
            out_append(elem)
        elif out[-1] is True and elem is True:
            out_append(True)
        elif out[-1] is False or elem is False:
            out_append(False)
        else:
            out_append(NA)
    return Series(out) if out else Series([], dtype=bool)


@register_func(None, context=Context.EVAL)
def cumany(series: Any) -> Series:
    """Get cumulative bool. All cases after first True"""
    if is_scalar(series):
        series = [series]

    # numpy treats NA differently than R does
    bools = boolean(series)  # all to bool, with NAs kept
    out = []
    out_append = out.append
    for elem in bools:
        if not out:
            out_append(elem)
        elif out[-1] is True or elem is True:
            out_append(True)
        else:
            out_append(NA)
    # let itself choose dtype (bool or object if NA exists)
    return Series(out) if out else Series([], dtype=bool)


@register_func(None, context=Context.EVAL)
def coalesce(x: Any, *replace: Any) -> Any:
    """Replace missing values

    https://dplyr.tidyverse.org/reference/coalesce.html

    Args:
        x: The vector to replace
        replace: The replacement

    Returns:
        A vector the same length as the first argument with missing values
        replaced by the first non-missing value.
    """
    if not replace:
        return x

    if isinstance(x, DataFrame):
        y = x.copy()
        for repl in replace:
            x = y.combine_first(repl)
            y = x
        return y

    if is_iterable(x):
        x = Series(x)
        for repl in replace:
            x = x.combine_first(
                Series(repl if is_iterable(repl) else [repl] * len(x))
            )
        return x.values

    return replace[0] if is_null(x) else x


@register_func(None, context=Context.EVAL)
def na_if(x: Iterable[Any], y: Any) -> Iterable[Any]:
    """Convert an annoying value to NA

    Args:
        x: Vector to modify
        y: Value to replace with NA

    Returns:
        A vector with values replaced.
    """
    if is_scalar(x):
        x = [x]
    if not isinstance(x, Series):
        x = Series(x, dtype=object) if len(x) == 0 else Series(x)

    x = x.copy()
    x[x == y] = NA
    return x


@register_func(None, context=Context.EVAL)
def near(x: Iterable[Any], y: Any, tol: float = 1e-8) -> Iterable[bool]:
    """Compare numbers with tolerance"""
    if is_scalar(x):
        x = [x]

    return numpy.isclose(x, y, atol=tol)


@register_func(None, context=Context.EVAL)
def nth(
    x: Iterable[Any],
    n: int,
    order_by: Iterable[Any] = None,
    default: Any = NA,
    base0_: bool = None,
) -> Any:
    """Get the nth element of x

    See https://dplyr.tidyverse.org/reference/nth.html

    Args:
        x: A collection of elements
        n: The order of the elements.
        order_by: An optional vector used to determine the order
        default: A default value to use if the position does not exist
            in the input.
        base0_: Whether `n` is 0-based or not.

    Returns:
        A single element of x at `n'th`
    """
    x = Array(x)
    if order_by is not None:
        order_by = Array(order_by)
        x = x[order_by.argsort()]
    if not isinstance(n, int):
        raise TypeError("`nth` expects `n` to be an integer")

    try:
        return x[position_at(n, len(x), base0=base0_)]
    except (ValueError, IndexError, TypeError):
        return default


@register_func(None, context=Context.EVAL)
def first(
    x: Iterable[Any],
    order_by: Iterable[Any] = None,
    default: Any = NA,
) -> Any:
    """Get the first element of x"""
    x = Array(x)
    if order_by is not None:
        order_by = Array(order_by)
        x = x[order_by.argsort()]
    try:
        return x[0]
    except IndexError:
        return default


@register_func(None, context=Context.EVAL)
def last(
    x: Iterable[Any],
    order_by: Iterable[Any] = None,
    default: Any = NA,
) -> Any:
    """Get the last element of x"""
    x = Array(x)
    if order_by is not None:
        order_by = Array(order_by)
        x = x[order_by.argsort()]
    try:
        return x[-1]
    except IndexError:
        return default


def boolean(value: Any) -> BoolOrIter:
    """Convert value to bool, but keep NAs"""
    if is_scalar(value):
        return NA if is_null(value) else bool(value)
    return [boolean(elem) for elem in value]



"""Provides functions for multiple factors"""
import itertools

from pandas import Categorical
from pipda.utils import CallingEnvs

from ..core.types import ForcatsType

from ..base import factor, paste, levels, expandgrid, intersect
from .utils import check_factor
from .lvls import lvls_union


def fct_c(*fs: ForcatsType) -> Categorical:
    """Concatenate factors, combining levels

    This is a useful way of patching together factors from multiple sources
    that really should have the same levels but don't.

    Args:
        fs: The factors to be concatenated

    Returns:
        The concatenated factor
    """
    if not fs:
        return factor()

    levs = lvls_union(fs)
    allvals = itertools.chain(*fs)
    return factor(allvals, levels=levs, exclude=None)


def fct_cross(
    *fs: ForcatsType,
    sep: str = ":",
    keep_empty: bool = False,
) -> Categorical:
    """Combine levels from two or more factors to create a new factor

    Computes a factor whose levels are all the combinations of
    the levels of the input factors.

    Args:
        *fs: Factors to combine
        sep: A string to separate levels
        keep_empty: If True, keep combinations with no observations as levels

    Returns:
        The new factor
    """
    if not fs:
        return factor()

    fs = [check_factor(fct) for fct in fs]
    newf = paste(*fs, sep=sep, __calling_env=CallingEnvs.REGULAR)

    old_levels = (levels(fct, __calling_env=CallingEnvs.REGULAR) for fct in fs)
    grid = expandgrid(*old_levels, __calling_env=CallingEnvs.REGULAR)
    new_levels = paste(
        *(grid[col] for col in grid),
        sep=sep,
        __calling_env=CallingEnvs.REGULAR,
    )

    if not keep_empty:
        new_levels = intersect(
            new_levels,
            newf,
            __calling_env=CallingEnvs.REGULAR,
        )

    return factor(newf, levels=new_levels)



"""Special mathematical functions related to the beta and gamma functions."""

import numpy
from pipda import register_func

from ..core.types import FloatOrIter, IntOrIter, NumericOrIter, is_scalar
from ..core.contexts import Context
from ..core.utils import Array
from .bessel import _get_special_func_from_scipy
from .na import NA

# beta(a, b)              => scipy.special.beta(a,b)
# lbeta(a, b)             => scipy.special.betaln(a,b)

# gamma(x)                => scipy.special.gamma(x)
# lgamma(x)               => scipy.special.gammaln(x)
# psigamma(x, deriv = 0)  => scipy.special.polygamma(deriv, x)
# digamma(x)              => scipy.special.digamma(x)
# trigamma(x)             => scipy.special.polygamma(1, x)

# choose(n, k)            => scipy.special.comb(n, k)
# lchoose(n, k)           => numpy.log(scipy.special.comb(n,k))
# factorial(x)            => scipy.special.factoral(x)
# lfactorial(x)           => numpy.log(scipy.special.factoral(x))


@register_func(None, context=Context.EVAL)
def beta(a: FloatOrIter, b: FloatOrIter) -> FloatOrIter:
    """The beta function

    Note that when both `a` and `b` are iterables, the broadcast mechanism is
    different from R:
        >>> a = [1,2]; b = [3, 4, 5] will turn into
        >>> [(1, 3), (2, 4), (1, 5)] in R but raises error in python

    Args:
        a: and
        b: Real-valued arguments

    Returns:
        Value of the beta function
    """
    fun = _get_special_func_from_scipy("beta")
    return fun(a, b)


@register_func(None, context=Context.EVAL)
def lbeta(a: FloatOrIter, b: FloatOrIter) -> FloatOrIter:
    """The natural logarithm of `beta()`

    Args:
        a: and
        b: Real-valued arguments

    Returns:
        The natural logarithm of value of the beta function
    """
    return numpy.log(beta(a, b))


@register_func(None, context=Context.EVAL)
def gamma(x: FloatOrIter) -> FloatOrIter:
    """The gamma function

    Args:
        x: Real or complex valued argument

    Returns:
        Values of the gamma function
    """
    fun = _get_special_func_from_scipy("gamma")
    if is_scalar(x) and x <= 0:
        return NA
    if not is_scalar(x):
        x = Array(x, dtype=float)
        x[x <= 0.0] = NA
    return fun(x)


@register_func(None, context=Context.EVAL)
def lgamma(x: FloatOrIter) -> FloatOrIter:
    """The nature logarithm of `gamma()`

    Args:
        x: Real or complex valued argument

    Returns:
        The natural logarithm of values of the gamma function
    """
    return numpy.log(gamma(x))


@register_func(None, context=Context.EVAL)
def digamma(x: FloatOrIter) -> FloatOrIter:
    """The digamma function.

    Args:
        x: Real or complex valued argument

    Returns:
        Computed values of psi.
    """
    fun = _get_special_func_from_scipy("digamma")
    if x == 0.0:
        return NA
    return fun(x)


@register_func(None, context=Context.EVAL)
def choose(n: FloatOrIter, k: IntOrIter) -> FloatOrIter:
    """The number of combinations of N things taken k at a time.

    Note that when both `a` and `b` are iterables, the broadcast mechanism is
    different from R:
        >>> a = [1,2]; b = [3, 4, 5] will turn into
        >>> [(1, 3), (2, 4), (1, 5)] in R but raises error in python

    Args:
        n: Number of things.
        k: Number of elements taken.

    Returns:
        The total number of combinations.
    """
    fun = _get_special_func_from_scipy("comb")
    return fun(n, k)


@register_func(None, context=Context.EVAL)
def lchoose(n: FloatOrIter, k: IntOrIter) -> FloatOrIter:
    """The natural logarithm of `choose()`

    Note that when both `a` and `b` are iterables, the broadcast mechanism is
    different from R:
        >>> a = [1,2]; b = [3, 4, 5] will turn into
        >>> [(1, 3), (2, 4), (1, 5)] in R but raises error in python

    Args:
        n: Number of things.
        k: Number of elements taken.

    Returns:
        The natural logarithm of the total number of combinations.
    """
    return numpy.log(choose(n, k))


@register_func(None, context=Context.EVAL)
def factorial(x: FloatOrIter) -> FloatOrIter:
    """The factorial of a number or array of numbers.

    Args:
        Input values. If x < 0, the return value is NA.

    Returns:
        Factorial of x
    """
    fun = _get_special_func_from_scipy("factorial")
    out = fun(x)

    if is_scalar(x):
        out = NA if x < 0 else out
        return out

    out[Array(x) < 0] = NA
    return out


@register_func(None, context=Context.EVAL)
def lfactorial(x: FloatOrIter) -> FloatOrIter:
    """The natural logarithm of `factorial()`

    Args:
        Input values. If x < 0, the return value is NA.

    Returns:
        The natural logarithm of factorial of x
    """
    return numpy.log(factorial(x))


@register_func(None, context=Context.EVAL)
def trigamma(x: FloatOrIter) -> FloatOrIter:
    """The second derivatives of the logarithm of the gamma function

    Args:
        x: A numeric value or iterable

    Returns:
        The value of the 2nd derivatives of the logarithm of the gamma function
    """
    fun = _get_special_func_from_scipy("polygamma")
    return fun(1, x)


@register_func(None, context=Context.EVAL)
def psigamma(x: FloatOrIter, deriv: NumericOrIter) -> FloatOrIter:
    """The deriv-th derivatives of the logarithm of the gamma function

    Args:
        x: A numeric value or iterable

    Returns:
        The value of the 2nd derivatives of the logarithm of the gamma function
    """
    fun = _get_special_func_from_scipy("polygamma")
    deriv = numpy.round(deriv)
    return fun(deriv, x)



"""Subset rows using their positions

https://github.com/tidyverse/dplyr/blob/master/R/slice.R
"""
import builtins
from typing import Any, Iterable, List, Union

from pandas import DataFrame, RangeIndex
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.collections import Collection
from ..core.utils import reconstruct_tibble
from ..core.grouped import DataFrameGroupBy, DataFrameRowwise
from ..core.types import NumericOrIter

from ..base import NA, unique
from .group_by import ungroup
from .dfilter import _filter_groups


@register_verb(DataFrame, context=Context.SELECT)
def slice(
    _data: DataFrame,
    *rows: NumericOrIter,
    _preserve: bool = False,
    base0_: bool = None,
) -> DataFrame:
    """Index rows by their (integer) locations

    Original APIs https://dplyr.tidyverse.org/reference/slice.html

    Args:
        _data: The dataframe
        rows: The indexes
            Ranges can be specified as `f[1:3]`
            Note that the negatives mean differently than in dplyr.
            In dplyr, negative numbers meaning exclusive, but here negative
            numbers are negative indexes like how they act in python indexing.
            For exclusive indexes, you need to use inversion. For example:
            `slice(df, ~f[:3])` excludes first 3 rows. You can also do:
            `slice(df, ~c(f[:3], 6))` to exclude multiple set of rows.
            To exclude a single row, you can't do this directly: `slice(df, ~1)`
            since `~1` is directly compiled into a number. You can do this
            instead: `slice(df, ~c(1))`
            Exclusive and inclusive expressions are allowed to be mixed, unlike
            in `dplyr`. They are expanded in the order they are passed in.
        _preserve: Relevant when the _data input is grouped.
            If _preserve = FALSE (the default), the grouping structure is
            recalculated based on the resulting data,
            otherwise the grouping is kept as is.
        base0_: If rows are selected by indexes, whether they are 0-based.
            If not provided, `datar.base.get_option('index.base.0')` is used.

    Returns:
        The sliced dataframe
    """
    if not rows:
        return _data

    rows = _sanitize_rows(rows, _data.shape[0], base0_)
    out = _data.iloc[rows, :]
    if isinstance(_data.index, RangeIndex):
        out.reset_index(drop=True, inplace=True)
    # copy_attrs(out, _data) # attrs carried
    return out


@slice.register(DataFrameGroupBy, context=Context.PENDING)
def _(
    _data: DataFrameGroupBy,
    *rows: Any,
    _preserve: bool = False,
    base0_: bool = None,
) -> DataFrameGroupBy:
    """Slice on grouped dataframe"""
    out = _data._datar_apply(
        slice, *rows, base0_=base0_, _drop_index=False
    ).sort_index()
    out = reconstruct_tibble(_data, out)
    gdata = _filter_groups(out, _data, sort_rows=False)

    if not _preserve and _data.attrs.get("_group_drop", True):
        out.attrs["_group_data"] = gdata[gdata["_rows"].map(len) > 0]

    return out


@slice.register(DataFrameRowwise, context=Context.PENDING)
def _(
    _data: DataFrameRowwise,
    *rows: Any,
    _preserve: bool = False,
    base0_: bool = None,
) -> DataFrameRowwise:
    """Slice on grouped dataframe"""
    out = slice.dispatch(DataFrame)(
        _data,
        *rows,
        _preserve=_preserve,
        base0_=base0_,
    )
    return reconstruct_tibble(_data, out, keep_rowwise=True)


@register_verb(DataFrame)
def slice_head(
    _data: DataFrame, n: int = None, prop: float = None
) -> DataFrame:
    """Select first rows

    Args:
        _data: The dataframe.
        n: and
        prop: Provide either n, the number of rows, or prop, the proportion of
            rows to select.
            If neither are supplied, n = 1 will be used.
            If n is greater than the number of rows in the group (or prop > 1),
            the result will be silently truncated to the group size.
            If the proportion of a group size is not an integer,
            it is rounded down.

    Returns:
        The sliced dataframe
    """
    n = _n_from_prop(_data.shape[0], n, prop)
    return slice(
        _data,
        builtins.slice(None, n),
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )


@slice_head.register(DataFrameGroupBy, context=Context.PENDING)
def _(_data: DataFrame, n: int = None, prop: float = None) -> DataFrameGroupBy:
    """Slice on grouped dataframe"""
    out = _data._datar_apply(
        slice_head,
        n=n,
        prop=prop,
        _drop_index=False,
        __calling_env=CallingEnvs.REGULAR,
    ).sort_index()
    return reconstruct_tibble(_data, out)


@slice_head.register(DataFrameRowwise, context=Context.PENDING)
def _(
    _data: DataFrameRowwise, n: int = None, prop: float = None
) -> DataFrameRowwise:
    """Slice on grouped dataframe"""
    out = slice_head.dispatch(DataFrame)(_data, n=n, prop=prop)
    return reconstruct_tibble(_data, out, keep_rowwise=True)


@register_verb(DataFrame)
def slice_tail(_data: DataFrame, n: int = 1, prop: float = None) -> DataFrame:
    """Select last rows

    See Also:
        [`slice_head()`](datar.dplyr.slice.slice_head)
    """
    n = _n_from_prop(_data.shape[0], n, prop)
    return slice(
        _data,
        builtins.slice(-n, None),
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )


@slice_tail.register(DataFrameGroupBy, context=Context.PENDING)
def _(_data: DataFrame, n: int = None, prop: float = None) -> DataFrameGroupBy:
    """Slice on grouped dataframe"""
    out = _data._datar_apply(
        slice_tail, n=n, prop=prop, _drop_index=False
    ).sort_index()
    return reconstruct_tibble(_data, out)


@slice_tail.register(DataFrameRowwise, context=Context.PENDING)
def _(
    _data: DataFrameRowwise, n: int = None, prop: float = None
) -> DataFrameRowwise:
    """Slice on grouped dataframe"""
    out = slice_tail.dispatch(DataFrame)(_data, n=n, prop=prop)
    return reconstruct_tibble(_data, out, keep_rowwise=True)


@register_verb(DataFrame, extra_contexts={"order_by": Context.EVAL})
def slice_min(
    _data: DataFrame,
    order_by: Iterable[Any],
    n: int = 1,
    prop: float = None,
    with_ties: Union[bool, str] = True,
) -> DataFrame:
    """select rows with lowest values of a variable.

    See Also:
        [`slice_head()`](datar.dplyr.slice.slice_head)
    """
    n = _n_from_prop(_data.shape[0], n, prop)

    sorting_df = DataFrame(index=_data.index)
    sorting_df["x"] = order_by
    keep = {True: "all", False: "first"}.get(with_ties, with_ties)
    sorting_df = sorting_df.nsmallest(n, "x", keep)

    out = _data.copy()  # attrs copied
    out = out.loc[sorting_df.index, :]  # attrs kept
    return out


@slice_min.register(DataFrameGroupBy, context=Context.PENDING)
def _(
    _data: DataFrameGroupBy,
    order_by: Iterable[Any],
    n: int = 1,
    prop: float = None,
    with_ties: Union[bool, str] = True,
) -> DataFrameGroupBy:
    """slice_min for DataFrameGroupBy object"""
    out = _data._datar_apply(
        slice_min,
        order_by=order_by,
        n=n,
        prop=prop,
        with_ties=with_ties,
    )
    return reconstruct_tibble(_data, out)


@slice_min.register(DataFrameRowwise, context=Context.EVAL)
def _(
    _data: DataFrameRowwise,
    order_by: Iterable[Any],
    n: int = 1,
    prop: float = None,
    with_ties: Union[bool, str] = True,
) -> DataFrameRowwise:
    """slice_min for DataFrameRowwise object"""
    out = slice_min.dispatch(DataFrame)(
        ungroup(_data, __calling_env=CallingEnvs.REGULAR),
        order_by=order_by,
        n=n,
        prop=prop,
        with_ties=with_ties,
    )
    return reconstruct_tibble(_data, out, keep_rowwise=True)


@register_verb(DataFrame, extra_contexts={"order_by": Context.EVAL})
def slice_max(
    _data: DataFrame,
    order_by: Iterable[Any],
    n: int = 1,
    prop: float = None,
    with_ties: Union[bool, str] = True,
) -> DataFrame:
    """select rows with highest values of a variable.

    See Also:
        [`slice_head()`](datar.dplyr.slice.slice_head)
    """
    n = _n_from_prop(_data.shape[0], n, prop)

    sorting_df = DataFrame(index=_data.index)
    sorting_df["x"] = order_by
    keep = {True: "all", False: "first"}.get(with_ties, with_ties)
    sorting_df = sorting_df.nlargest(n, "x", keep)

    out = _data.copy()  # attrs copied
    out = out.loc[sorting_df.index, :]  # attrs kept
    return out


@slice_max.register(DataFrameGroupBy, context=Context.PENDING)
def _(
    _data: DataFrameGroupBy,
    order_by: Iterable[Any],
    n: int = 1,
    prop: float = None,
    with_ties: Union[bool, str] = True,
) -> DataFrameGroupBy:
    """slice_max for DataFrameGroupBy object"""
    out = _data._datar_apply(
        slice_max,
        order_by=order_by,
        n=n,
        prop=prop,
        with_ties=with_ties,
    )
    return reconstruct_tibble(_data, out)


@slice_max.register(DataFrameRowwise, context=Context.EVAL)
def _(
    _data: DataFrameRowwise,
    order_by: Iterable[Any],
    n: int = 1,
    prop: float = None,
    with_ties: Union[bool, str] = True,
) -> DataFrameRowwise:
    """slice_max for DataFrameRowwise object"""
    out = slice_max.dispatch(DataFrame)(
        ungroup(_data, __calling_env=CallingEnvs.REGULAR),
        order_by=order_by,
        n=n,
        prop=prop,
        with_ties=with_ties,
    )
    return reconstruct_tibble(_data, out, keep_rowwise=True)


@register_verb(DataFrame, extra_contexts={"weight_by": Context.EVAL})
def slice_sample(
    _data: DataFrame,
    n: int = 1,
    prop: float = None,
    weight_by: Iterable[Union[int, float]] = None,
    replace: bool = False,
    random_state: Any = None,
) -> DataFrame:
    """Randomly selects rows.

    See Also:
        [`slice_head()`](datar.dplyr.slice.slice_head)
    """
    n = _n_from_prop(_data.shape[0], n, prop)
    if n == 0:
        # otherwise _data.sample raises error when weight_by is empty as well
        return _data.iloc[[], :]

    return _data.sample(
        n=n,
        replace=replace,
        weights=weight_by,
        random_state=random_state,
        axis=0,
    )


@slice_sample.register(DataFrameGroupBy, context=Context.PENDING)
def _(
    _data: DataFrameGroupBy,
    n: int = 1,
    prop: float = None,
    weight_by: Iterable[Union[int, float]] = None,
    replace: bool = False,
    random_state: Any = None,
) -> DataFrameGroupBy:
    out = _data._datar_apply(
        slice_sample,
        n=n,
        prop=prop,
        weight_by=weight_by,
        replace=replace,
        random_state=random_state,
    )
    return reconstruct_tibble(_data, out)


@slice_sample.register(DataFrameRowwise, context=Context.EVAL)
def _(
    _data: DataFrameRowwise,
    n: int = 1,
    prop: float = None,
    weight_by: Iterable[Union[int, float]] = None,
    replace: bool = False,
    random_state: Any = None,
) -> DataFrameRowwise:
    out = slice_sample.dispatch(DataFrame)(
        _data,
        n=n,
        prop=prop,
        weight_by=weight_by,
        replace=replace,
        random_state=random_state,
    )
    return reconstruct_tibble(_data, out, keep_rowwise=True)


def _n_from_prop(
    total: int,
    n: int = None,
    prop: float = None,
) -> int:
    """Get n from a proportion"""
    if n is None and prop is None:
        return 1
    if n is not None and not isinstance(n, int):
        raise TypeError(f"Expect `n` int type, got {type(n)}.")
    if prop is not None and not isinstance(prop, (int, float)):
        raise TypeError(f"Expect `prop` a number, got {type(n)}.")
    if (n is not None and n < 0) or (prop is not None and prop < 0):
        raise ValueError("`n` and `prop` should not be negative.")
    if prop is not None:
        return int(float(total) * min(prop, 1.0))
    return min(n, total)


def _sanitize_rows(rows: Iterable, nrow: int, base0: bool = None) -> List[int]:
    """Sanitize rows passed to slice"""
    rows = Collection(*rows, pool=nrow, base0=base0)
    if rows.error:

        raise rows.error from None
    invalid_type_rows = [
        row
        for row in rows.unmatched
        if not isinstance(row, (int, type(None), type(NA)))
    ]
    if invalid_type_rows:
        raise TypeError("`slice()` expressions should return indices.")
    return unique(rows)



"""Subset distinct/unique rows

See source https://github.com/tidyverse/dplyr/blob/master/R/distinct.R
"""
from typing import Any

import pandas
from pandas import DataFrame
from pipda import register_verb, register_func
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.utils import copy_attrs, reconstruct_tibble
from ..core.grouped import DataFrameGroupBy, DataFrameRowwise
from ..core.types import is_scalar
from ..core.collections import Collection
from ..base import union, setdiff, intersect
from .mutate import mutate
from .group_by import ungroup


@register_verb(DataFrame, context=Context.PENDING)
def distinct(
    _data: DataFrame, *args: Any, _keep_all: bool = False, **kwargs: Any
) -> DataFrame:
    """Select only unique/distinct rows from a data frame.

    The original API:
    https://dplyr.tidyverse.org/reference/distinct.html

    Args:
        _data: The dataframe
        *args: and
        **kwargs: Optional variables to use when determining
            uniqueness.
        _keep_all: If TRUE, keep all variables in _data

    Returns:
        A dataframe without duplicated rows in _data
    """
    if not args and not kwargs:
        uniq = _data.drop_duplicates()
    else:
        # keep_none_prefers_new_order
        uniq = (
            mutate(
                _data,
                *args,
                **kwargs,
                _keep="none",
                __calling_env=CallingEnvs.REGULAR,
            )
        ).drop_duplicates()

    if not _keep_all:
        # keep original order
        out = uniq[
            union(
                intersect(
                    _data.columns,
                    uniq.columns,
                    __calling_env=CallingEnvs.REGULAR,
                ),
                setdiff(
                    uniq.columns,
                    _data.columns,
                    __calling_env=CallingEnvs.REGULAR,
                ),
                __calling_env=CallingEnvs.REGULAR,
            )
        ]
    else:
        out = _data.loc[uniq.index, :].copy()
        out[uniq.columns.tolist()] = uniq
    copy_attrs(out, _data)
    return out.reset_index(drop=True)


@distinct.register(DataFrameGroupBy, context=Context.PENDING)
def _(
    _data: DataFrameGroupBy, *args: Any, _keep_all: bool = False, **kwargs: Any
) -> DataFrameGroupBy:

    out = _data._datar_apply(
        lambda df: distinct(
            df,
            *args,
            **kwargs,
            _keep_all=_keep_all,
            __calling_env=CallingEnvs.REGULAR,
        )
    )

    return reconstruct_tibble(_data, out)


@distinct.register(DataFrameRowwise, context=Context.PENDING)
def _(
    _data: DataFrameRowwise, *args: Any, _keep_all: bool = False, **kwargs: Any
) -> DataFrameRowwise:

    out = distinct.dispatch(DataFrame)(
        ungroup(_data, __calling_env=CallingEnvs.REGULAR),
        *args,
        **kwargs,
        _keep_all=_keep_all,
    )

    return reconstruct_tibble(_data, out, keep_rowwise=True)


@register_func(None, context=Context.EVAL)
def n_distinct(*data: Any, na_rm: bool = False) -> int:
    """Get the length of distince elements"""
    data = Collection(*(dat if is_scalar(dat) else list(dat) for dat in data))
    data = pandas.unique(data)
    if not na_rm:
        return len(data)
    return pandas.notna(data).sum()



"""Fill in missing values with previous or next value

https://github.com/tidyverse/tidyr/blob/HEAD/R/fill.R
"""

from typing import Union

from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.utils import vars_select, reconstruct_tibble
from ..core.grouped import DataFrameGroupBy


@register_verb(DataFrame, context=Context.SELECT)
def fill(
    _data: DataFrame,
    *columns: Union[str, int],
    _direction: str = "down",
    base0_: bool = None,
) -> DataFrame:
    """Fills missing values in selected columns using the next or
    previous entry.

    See https://tidyr.tidyverse.org/reference/fill.html

    Args:
        _data: A dataframe
        *columns: Columns to fill
        _direction: Direction in which to fill missing values.
            Currently either "down" (the default), "up",
            "downup" (i.e. first down and then up) or
            "updown" (first up and then down).
        base0_: Whether `*columns` are 0-based if given by indexes
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The dataframe with NAs being replaced.
    """
    data = _data.copy()
    if not columns:
        data = data.fillna(
            method="ffill" if _direction.startswith("down") else "bfill",
        )
        if _direction in ("updown", "downup"):
            data = data.fillna(
                method="ffill" if _direction.endswith("down") else "bfill",
            )
    else:
        colidx = vars_select(data.columns, *columns, base0=base0_)
        data.iloc[:, colidx] = fill(
            data.iloc[:, colidx],
            _direction=_direction,
            __calling_env=CallingEnvs.REGULAR,
        )
    return data


@fill.register(DataFrameGroupBy, context=Context.SELECT)
def _(
    _data: DataFrameGroupBy, *columns: str, _direction: str = "down"
) -> DataFrameGroupBy:
    # DataFrameGroupBy
    out = _data._datar_apply(
        fill,
        *columns,
        _direction=_direction,
        _drop_index=False,
    ).sort_index()
    return reconstruct_tibble(_data, out, keep_rowwise=True)



"""String functions in R"""
import re
import warnings
from typing import Any, Iterable, Tuple, Union

import numpy
from pandas.core.dtypes.common import is_string_dtype
from pipda import register_func

from ..core.contexts import Context
from ..core.types import (
    BoolOrIter,
    Dtype,
    IntOrIter,
    StringOrIter,
    is_null,
    is_scalar
)
from ..core.utils import (
    Array,
    arg_match,
    get_option,
    length_of,
    logger,
    position_at,
    recycle_value
)
from .casting import _as_type
from .na import NA
from .testing import _register_type_testing


@register_func(None, context=Context.EVAL)
def as_character(x: Any, str_dtype: Dtype = str, _na: Any = NA) -> StringOrIter:
    """Convert an object or elements of an iterable into string

    Aliases `as_str` and `as_string`

    Args:
        x: The object
        str_dtype: The string dtype to convert to
        _na: How NAs should be casted. Specify NA will keep them unchanged.
            But the dtype will be object then.

    Returns:
        When x is an array or a series, return x.astype(str).
        When x is iterable, convert elements of it into strings
        Otherwise, convert x to string.
    """
    return _as_type(x, str_dtype, na=_na)


as_str = as_string = as_character

is_character = _register_type_testing(
    "is_character",
    scalar_types=(str, numpy.str_),
    dtype_checker=is_string_dtype,
    doc="""Test if a value is characters/string

Alias `is_str` and `is_string`

Args:
    x: The value to be checked

Returns:
    True if the value is string; with a string dtype;
    or all elements are strings
""",
)

is_str = is_string = is_character


# Grep family -----------------------------------


@register_func(None, context=Context.EVAL)
def grep(
    pattern: StringOrIter,
    x: StringOrIter,
    ignore_case: bool = False,
    value: bool = False,
    fixed: bool = False,
    invert: bool = False,
    base0_: bool = None,
) -> Iterable[Union[int, str]]:
    """R's grep, get the element in x matching the pattern

    Args:
        pattern: The pattern
        x: A string or an iterable of strings; or those can be coerced to
        ignore_case: Do case-insensitive matching?
        value: Return values instead of indices?
        fixed: Fixed matching (instead of regex matching)?
        invert: Return elements thata don't match instead?
        base0_: When return indices, whether return 0-based indices?
            If not set, will use `datar.base.get_option('which.base.0')`

    Returns:
        The matched (or unmatched (`invert=True`)) indices
        (or values (`value=True`)).
    """
    if is_scalar(x):
        x = [x]  # type: ignore
    x = Array(as_character(x), dtype=object)
    matched = grepl(
        pattern=pattern,
        x=x,
        ignore_case=ignore_case,
        fixed=fixed,
        invert=invert,
    )

    if value:
        return x[matched]

    base0_ = get_option("which.base.0", base0_)
    return numpy.flatnonzero(matched) + int(not base0_)


@register_func(None, context=Context.EVAL)
def grepl(
    pattern: StringOrIter,
    x: StringOrIter,
    ignore_case: bool = False,
    fixed: bool = False,
    invert: bool = False,
) -> Iterable[Union[int, str]]:
    """R's grepl, check whether elements in x matching the pattern

    Args:
        pattern: The pattern
        x: A string or an iterable of strings; or those can be coerced to
        ignore_case: Do case-insensitive matching?
        fixed: Fixed matching (instead of regex matching)?
        invert: Return elements thata don't match instead?

    Returns:
        A bool array indicating whether the elements in x match the pattern
    """
    pattern = _warn_more_pat_or_rep(pattern, "grep")
    match_fun = lambda text: _match(
        pattern, text, ignore_case=ignore_case, invert=invert, fixed=fixed
    )
    if is_scalar(x):
        x = [x]  # type: ignore
    x = Array(as_character(x), dtype=object)
    return Array(list(map(match_fun, x)), dtype=bool)


@register_func(None, context=Context.EVAL)
def sub(
    pattern: StringOrIter,
    replacement: StringOrIter,
    x: StringOrIter,
    ignore_case: bool = False,
    fixed: bool = False,
) -> Iterable[str]:
    """R's sub, replace a pattern with replacement for elements in x,
    each only once

    Args:
        pattern: The pattern
        replacement: The replacement
        x: A string or an iterable of strings; or those can be coerced to
        ignore_case: Do case-insensitive matching?
        fixed: Fixed matching (instead of regex matching)?

    Returns:
        An array of strings with matched parts replaced.
    """
    return _sub(
        pattern=pattern,
        replacement=replacement,
        x=x,
        ignore_case=ignore_case,
        fixed=fixed,
    )


@register_func(None, context=Context.EVAL)
def gsub(
    pattern: StringOrIter,
    replacement: StringOrIter,
    x: StringOrIter,
    ignore_case: bool = False,
    fixed: bool = False,
) -> Iterable[str]:
    """R's gsub, replace a pattern with replacement for elements in x,
    each for all matched parts

    See Also:
        [sub()](datar.base.string.sub)
    """
    return _sub(
        pattern=pattern,
        replacement=replacement,
        x=x,
        ignore_case=ignore_case,
        fixed=fixed,
        count=0,
        fun="gsub",
    )


# Grep family helpers --------------------------------


def _warn_more_pat_or_rep(
    pattern: StringOrIter, fun: str, arg: str = "pattern"
) -> str:
    """Warn when there are more than one pattern or replacement provided"""
    if is_scalar(pattern):
        return pattern
    if len(pattern) == 1:
        return pattern[0]

    logger.warning(
        "In %s(...), argument `%s` has length > 1 and only the "
        "first element will be used",
        fun,
        arg,
    )
    return pattern[0]


def _match(
    pattern: str, text: str, ignore_case: bool, invert: bool, fixed: bool
) -> bool:
    """Do the regex match"""
    if is_null(text):
        return False

    flags = re.IGNORECASE if ignore_case else 0
    if fixed:
        pattern = re.escape(pattern)

    pattern = re.compile(pattern, flags)
    matched = pattern.search(text)
    if invert:
        matched = not bool(matched)
    return bool(matched)


def _sub(
    pattern: StringOrIter,
    replacement: StringOrIter,
    x: StringOrIter,
    ignore_case: bool = False,
    fixed: bool = False,
    count: int = 1,
    fun: str = "sub",
) -> Iterable[str]:
    """Replace a pattern with replacement for elements in x,
    with argument count available
    """
    pattern = _warn_more_pat_or_rep(pattern, fun)
    replacement = _warn_more_pat_or_rep(replacement, fun, "replacement")
    if fixed:
        pattern = re.escape(pattern)

    flags = re.IGNORECASE if ignore_case else 0
    pattern = re.compile(pattern, flags)

    if is_scalar(x):
        return pattern.sub(repl=replacement, count=count, string=x)

    return Array(
        [pattern.sub(repl=replacement, count=count, string=elem) for elem in x],
        dtype=object,
    )


@register_func(None, context=Context.EVAL)
def nchar(
    x: StringOrIter,
    type: str = "chars",
    allow_na: bool = True,  # i.e.: '\ud861'
    keep_na: bool = None,
    _na_len: int = 2,
) -> Iterable[int]:
    """Get the size of the elements in x"""
    x, keep_na = _prepare_nchar(x, type, keep_na)
    out = [
        _nchar_scalar(
            elem, retn=type, allow_na=allow_na, keep_na=keep_na, na_len=_na_len
        )
        for elem in x
    ]
    if is_null(out).any():
        return Array(out, dtype=object)
    return Array(out, dtype=int)


@register_func(None, context=Context.EVAL)
def nzchar(x: StringOrIter, keep_na: bool = False) -> Iterable[bool]:
    """Find out if elements of a character vector are non-empty strings.

    Args:
        x: Strings to test
        keep_na: Keep NAs as is?

    Returns:
        A bool array to tell whether elements in x are non-empty strings
    """
    if is_scalar(x):
        x = [x]  # type: ignore
    x = as_character(x, _na=NA if keep_na else None)
    out = [NA if is_null(elem) else bool(elem) for elem in x]
    if is_null(out).any():
        return Array(out, dtype=object)
    return Array(out, dtype=bool)


# nchar helpers --------------------------------


def _prepare_nchar(
    x: StringOrIter, type: str, keep_na: bool
) -> Tuple[Iterable[str], bool]:
    """Prepare arguments for n(z)char"""
    arg_match(type, "type", ["chars", "bytes", "width"])
    if keep_na is None:
        keep_na = type != "width"

    if is_scalar(x):
        x = [x]  # type: ignore
    x = Array(as_character(x), dtype=object)
    return x, keep_na


def _nchar_scalar(
    x: str, retn: str, allow_na: bool, keep_na: bool, na_len: int
) -> int:
    """Get the size of a scalar string"""
    if is_null(x):
        return NA if keep_na else na_len

    if retn == "width":
        from wcwidth import wcswidth

        return wcswidth(x)
    if retn == "chars":
        return len(x)

    try:
        x = x.encode("utf-8")
    except UnicodeEncodeError:
        if allow_na:
            return NA
        raise
    return len(x)


# paste and paste0 --------------------


@register_func(None, context=Context.EVAL)
def paste(
    *args: StringOrIter, sep: str = " ", collapse: str = None
) -> StringOrIter:
    """Concatenate vectors after converting to character.

    Args:
        *args: strings to be concatenated
        sep: The separator
        collapse: The separator to collapse the final string arrays

    Returns:
        A single string if collapse is given, otherwise an array of strings.
    """
    args = [as_character(arg, _na="NA") for arg in args]
    maxlen = max(map(length_of, args))
    args = zip(*(recycle_value(arg, maxlen) for arg in args))

    out = [sep.join(arg) for arg in args]
    if collapse is not None:
        return collapse.join(out)

    return Array(out, dtype=str)


@register_func(None, context=Context.EVAL)
def paste0(*args: StringOrIter, collapse: str = None) -> StringOrIter:
    """`paste` with `sep=""`

    See [paste](datar.base.string.paste)
    """
    return paste(*args, sep="", collapse=collapse)


# sprintf ----------------------------------------------------------------


@register_func(None, context=Context.EVAL)
def sprintf(fmt: StringOrIter, *args: StringOrIter) -> StringOrIter:
    """C-style String Formatting

    Args:
        fmt: The formats
        *args: The values

    Returns:
        A scalar string if all fmt, *args are scalar strings, otherwise
        an array of formatted strings
    """
    if is_scalar(fmt) and all(is_scalar(arg) for arg in args):
        return fmt % args

    # args = [as_character(arg, _na='NA') for arg in args]
    maxlen = max(map(length_of, args))
    maxlen = max(maxlen, max(map(length_of, fmt)))

    fmt = recycle_value(fmt, maxlen)
    args = [recycle_value(arg, maxlen) for arg in args]
    return Array([fmt[i] % arg for i, arg in enumerate(zip(*args))], dtype=str)


# substr, substring ----------------------------------


@register_func(None, context=Context.EVAL)
def substr(
    x: StringOrIter,
    start: IntOrIter,
    stop: IntOrIter,
    base0_: bool = None,
) -> StringOrIter:
    """Extract substrings in strings.

    Args:
        x: The strings
        start: The start positions to extract
        stop: The stop positions to extract
        base0_: Whether `start` and `stop` are 0-based
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The substrings from `x`
    """
    if is_scalar(x) and is_scalar(start) and is_scalar(stop):
        if is_null(x):
            return NA
        base0_ = get_option("index.base.0", base0_)
        x = as_character(x)
        lenx = len(x)
        # int() converts numpy.int64 to int
        start0 = position_at(int(start), lenx, base0=base0_)
        stop0 = position_at(
            min(int(stop), lenx - int(base0_)), lenx, base0=base0_
        )
        return x[start0 : stop0 + 1]

    if is_scalar(x):
        x = [x]  # type: ignore
    if is_scalar(start):
        start = [start]  # type: ignore
    if is_scalar(stop):
        stop = [stop]  # type: ignore
    maxlen = max(length_of(x), length_of(start), length_of(stop))
    x = recycle_value(x, maxlen)
    start = recycle_value(start, maxlen)
    stop = recycle_value(stop, maxlen)
    out = [
        substr(elem, start_, stop_, base0_)
        for elem, start_, stop_ in zip(x, start, stop)
    ]
    if is_null(out).any():
        return Array(out, dtype=object)
    return Array(out, dtype=str)


@register_func(None, context=Context.EVAL)
def substring(
    x: StringOrIter,
    first: IntOrIter,
    last: IntOrIter = 1000000,
    base0_: bool = None,
) -> StringOrIter:
    """Extract substrings in strings.

    Args:
        x: The strings
        start: The start positions to extract
        stop: The stop positions to extract
        base0_: Whether `start` and `stop` are 0-based
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The substrings from `x`
    """
    return substr(x, first, last, base0_)


# strsplit --------------------------------


@register_func(None, context=Context.EVAL)
def strsplit(
    x: StringOrIter, split: StringOrIter, fixed: bool = False
) -> Iterable[Union[str, Iterable[str]]]:
    """Split strings by separator

    Args:
        x: The strings. Have to be strings, no casting will be done.
        split: The separators to split
        fixed: fixed matching (instead of regex matching)?

    Returns:
        List of split strings of x if both x and split are scalars. Otherwise,
        an array of split strings
    """
    if is_scalar(x) and is_scalar(split):
        if fixed:
            split = re.escape(split)
        split = re.compile(split)
        return split.split(x)

    maxlen = max(length_of(x), length_of(split))
    x = recycle_value(x, maxlen)
    split = recycle_value(split, maxlen)
    out = [strsplit(elem, splt, fixed=fixed) for elem, splt in zip(x, split)]
    return Array(out, dtype=object)


# startsWith, endsWith
@register_func(None, context=Context.EVAL)
def startswith(x: StringOrIter, prefix: StringOrIter) -> BoolOrIter:
    """Determines if entries of x start with prefix

    Args:
        x: A vector of strings or a string
        prefix: The prefix to test against

    Returns:
        A bool vector for each element in x if element startswith the prefix
    """
    if is_scalar(x):
        x = [x]

    prefix = recycle_value(prefix, len(x))
    # NAs?
    return Array(
        [elem.startswith(pref) for elem, pref in zip(x, prefix)],
        dtype=bool
    )


@register_func(None, context=Context.EVAL)
def endswith(x: StringOrIter, suffix: StringOrIter) -> BoolOrIter:
    """Determines if entries of x end with suffix

    Args:
        x: A vector of strings or a string
        suffix: The suffix to test against

    Returns:
        A bool vector for each element in x if element endswith the suffix
    """
    if is_scalar(x):
        x = [x]

    suffix = recycle_value(suffix, len(x))
    # NAs?
    return Array(
        [elem.endswith(suf) for elem, suf in zip(x, suffix)],
        dtype=bool
    )


@register_func(None, context=Context.EVAL)
def strtoi(x: StringOrIter, base: int = 0):
    """Convert strings to integers according to the given base

    Args:
        x: A string or vector of strings
        base: an integer which is between 2 and 36 inclusive, or zero.
            With zero, a suitable base will be chosen following the C rules.

    Returns:
        Converted integers
    """
    if is_scalar(x):
        return int(x, base=base)

    return Array([int(elem, base=base) for elem in x], dtype=int)


@register_func(None, context=Context.EVAL)
def chartr(old: StringOrIter, new: StringOrIter, x: Any) -> StringOrIter:
    """Replace strings char by char

    Args:
        old: A set of characters to replace
        new: A set of characters to replace with
        x: A string or vector of strings

    Returns:
        The strings in x being replaced
    """
    if len(old) > len(new):
        raise ValueError("'old' is longer than 'new'")

    if not is_scalar(old):
        old = old[0]
        warnings.warn(
            "argument 'old' has length > 1 and "
            "only the first element will be used"
        )
    if not is_scalar(new):
        new = new[0]
        warnings.warn(
            "argument 'new' has length > 1 and "
            "only the first element will be used"
        )

    new = new[:len(old)]

    def replace_single(elem: str) -> str:
        """Replace a single string"""
        out = elem[:]
        for oldc, newc in zip(old, new):
            out = out.replace(oldc, newc)
        return out

    if is_scalar(x):
        return replace_single(x)

    return Array([
        replace_single(elem) for elem in as_character(x)
    ])


@register_func(None, context=Context.EVAL)
def tolower(x: StringOrIter) -> StringOrIter:
    """Convert strings to lower case

    Args:
        x: A string or vector of strings

    Returns:
        Converted strings
    """
    x = as_character(x)
    if is_scalar(x):
        return x.lower()

    return Array([elem.lower() for elem in x])


@register_func(None, context=Context.EVAL)
def toupper(x: StringOrIter) -> StringOrIter:
    """Convert strings to upper case

    Args:
        x: A string or vector of strings

    Returns:
        Converted strings
    """
    x = as_character(x)
    if is_scalar(x):
        return x.upper()

    return Array([elem.upper() for elem in x])


@register_func(None, context=Context.EVAL)
def trimws(
    x: StringOrIter,
    which: str = "both",
    whitespace: str = r"[ \t\r\n]"
) -> StringOrIter:
    """Remove leading and/or trailing whitespace from character strings.

    Args:
        x: A string or vector of strings
        which: A character string specifying whether to remove
            both leading and trailing whitespace (default),
            or only leading ("left") or trailing ("right").
        whitespace: a string specifying a regular expression to
            match (one character of) “white space”

    Returns:
        The strings with whitespaces removed
    """
    which = arg_match(which, "which", ["both", "left", "right"])
    x = as_character(x)

    def trimws_single(elem: str) -> str:
        """Trim a single string"""
        if which == "both":
            expr = f"^{whitespace}|{whitespace}$"
        elif which == "left":
            expr = f"^{whitespace}"
        else:
            expr = f"{whitespace}$"

        return re.sub(expr, "", elem)

    if is_scalar(x):
        return trimws_single(x)

    return Array([trimws_single(elem) for elem in x])



"""Bind multiple data frames by row and column

See https://github.com/tidyverse/dplyr/blob/master/R/bind.r
"""
from typing import Callable, Union

import pandas
from pandas import DataFrame, Categorical
from pandas.api.types import union_categoricals
from pipda import register_verb

from ..core.types import NoneType, is_null, is_categorical
from ..core.contexts import Context
from ..core.utils import get_option, logger, reconstruct_tibble
from ..core.names import repair_names
from ..core.grouped import DataFrameGroupBy
from ..tibble import tibble


@register_verb((DataFrame, list, dict, NoneType), context=Context.EVAL)
def bind_rows(
    _data: Union[DataFrame, list, dict],
    *datas: Union[DataFrame, dict],
    _id: str = None,
    base0_: bool = None,
    _copy: bool = True,
    **kwargs: Union[DataFrame, dict],
) -> DataFrame:

    """Bind rows of give dataframes

    Original APIs https://dplyr.tidyverse.org/reference/bind.html

    Args:
        _data: The seed dataframe to bind others
            Could be a dict or a list, keys/indexes will be used for _id col
        *datas: Other dataframes to combine
        _id: The name of the id columns
        base0_: Whether `_id` starts from 0 or not, if no keys are provided.
            If `base0_` is not provided, will use
            `datar.base.get_option('index.base.0')`
        _copy: If `False`, do not copy data unnecessarily.
            Original API does not support this. This argument will be
            passed by to `pandas.concat()` as `copy` argument.
        **kwargs: A mapping of dataframe, keys will be used as _id col.

    Returns:
        The combined dataframe
    """
    base = int(not get_option("index.base.0", base0_))

    if _id is not None and not isinstance(_id, str):
        raise ValueError("`_id` must be a scalar string.")

    def data_to_df(data):
        """Make a copy of dataframe or convert dict to a dataframe"""
        if isinstance(data, DataFrame):
            return data.copy()

        out = tibble(**data)  # avoid varname error
        return out

    key_data = {}
    if isinstance(_data, list):
        for i, dat in enumerate(_data):
            if dat is not None:
                key_data[i + base] = data_to_df(dat)
    elif _data is not None:
        key_data[base] = data_to_df(_data)

    for i, dat in enumerate(datas):
        if isinstance(dat, list):
            for df in dat:
                key_data[len(key_data) + base] = data_to_df(df)
        elif dat is not None:
            key_data[len(key_data) + base] = data_to_df(dat)

    for key, val in kwargs.items():
        if val is not None:
            key_data[key] = data_to_df(val)

    if not key_data:
        return DataFrame()

    # handle categorical data
    for col in list(key_data.values())[0].columns:
        all_series = [
            dat[col]
            for dat in key_data.values()
            if col in dat and not dat[col].isna().all()
        ]
        all_categorical = [
            is_categorical(ser) or all(is_null(ser)) for ser in all_series
        ]
        if all(all_categorical):
            union_cat = union_categoricals(all_series)
            for data in key_data.values():
                if col not in data:  # in case it is 0-column df
                    continue
                data[col] = Categorical(
                    data[col],
                    categories=union_cat.categories,
                    ordered=is_categorical(data[col]) and data[col].cat.ordered,
                )
        elif any(all_categorical):
            logger.warning("Factor information lost during rows binding.")

    if _id is not None:
        return (
            pandas.concat(
                key_data.values(),
                keys=key_data.keys(),
                names=[_id, None],
                copy=_copy,
            )
            .reset_index(level=0)
            .reset_index(drop=True)
        )

    return pandas.concat(key_data.values(), copy=_copy).reset_index(drop=True)


@bind_rows.register(DataFrameGroupBy, context=Context.PENDING)
def _(
    _data: DataFrameGroupBy,
    *datas: Union[DataFrame, dict],
    _id: str = None,
    **kwargs: Union[DataFrame, dict],
) -> DataFrameGroupBy:

    data = bind_rows.dispatch(DataFrame)(_data, *datas, _id=_id, **kwargs)
    return reconstruct_tibble(_data, data)


@register_verb((DataFrame, dict, NoneType), context=Context.EVAL)
def bind_cols(
    _data: Union[DataFrame, dict],
    *datas: Union[DataFrame, dict],
    _name_repair: Union[str, Callable] = "unique",
    base0_: bool = None,
    _copy: bool = True,
) -> DataFrame:
    """Bind columns of give dataframes

    Note that unlike `dplyr`, mismatched dimensions are allowed and
    missing rows will be filled with `NA`s

    Args:
        _data: The seed dataframe to bind others
            Could be a dict, keys will be used for _id col
        *datas: other dataframes to bind
        _name_repair: treatment of problematic column names:
            - "minimal": No name repair or checks, beyond basic existence,
            - "unique": Make sure names are unique and not empty,
            - "check_unique": (default value), no name repair,
                but check they are unique,
            - "universal": Make the names unique and syntactic
            - a function: apply custom name repair
        base0_: Whether the numeric suffix starts from 0 or not.
            If not specified, will use `datar.base.get_option('index.base.0')`.
        _copy: If `False`, do not copy data unnecessarily.
            Original API does not support this. This argument will be
            passed by to `pandas.concat()` as `copy` argument.

    Returns:
        The combined dataframe
    """
    if isinstance(_data, dict):
        _data = tibble(**_data)
    more_data = []
    for data in datas:
        if isinstance(data, dict):
            more_data.append(tibble(**data))
        else:
            more_data.append(data)
    if _data is not None:
        more_data.insert(0, _data)
    if not more_data:
        return DataFrame()
    ret = pandas.concat(more_data, axis=1, copy=_copy)
    ret.columns = repair_names(
        ret.columns.tolist(), repair=_name_repair, base0_=base0_
    )
    return ret
